# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .conversion import *
from .model import utils
import numpy as np
import importlib
import argparse
import json
import six


def get_reader(input_format, decomposed, fold_constants, custom_shapes):
    if input_format == 'tf':
        from .io.tf.graphdef import Reader
        return Reader(fold_constants=fold_constants)
    elif input_format == 'tflite':
        from .io.tf.lite import Reader
        return Reader()
    elif input_format == 'nnef':
        from .io.nnef import Reader
        return Reader(decomposed=decomposed, custom_shapes=custom_shapes)
    elif input_format == 'onnx':
        from .io.onnx import Reader
        return Reader(simplify=fold_constants)
    elif input_format == 'caffe2':
        from .io.caffe2 import Reader
        return Reader()
    elif input_format == 'caffe':
        from .io.caffe2 import Reader
        return Reader(legacy=True)
    else:
        return None


def get_writer(output_format, fragments, fragment_dependencies, generate_fragments, annotate_shapes, compression):
    if output_format == 'tf':
        from .io.tf.graphdef import Writer
        return Writer()
    elif output_format == 'tflite':
        from .io.tf.lite import Writer
        return Writer()
    elif output_format == 'nnef':
        from .io.nnef import Writer
        return Writer(fragments=fragments, fragment_dependencies=fragment_dependencies,
                      generate_custom_fragments=generate_fragments,
                      annotate_shapes=annotate_shapes, compression=compression)
    elif output_format == 'onnx':
        from .io.onnx import Writer
        return Writer()
    elif output_format == 'caffe2':
        from .io.caffe2 import Writer
        return Writer()
    else:
        return None


def get_converter(input_format, output_format, io_transpose, custom_transforms, custom_functions, custom_shapes,
                  mirror_unsupported, keep_io_names):
    if input_format == 'tf' and output_format == 'nnef':
        from .conversion.tf_to_nnef import Converter
        return Converter(io_transpose=io_transpose,
                         custom_transforms=custom_transforms,
                         custom_functions=custom_functions,
                         mirror_unsupported=mirror_unsupported,
                         keep_io_names=keep_io_names)
    elif input_format == 'nnef' and output_format == 'tf':
        from .conversion.nnef_to_tf import Converter
        return Converter(io_transpose=io_transpose,
                         custom_transforms=custom_transforms,
                         custom_functions=custom_functions,
                         mirror_unsupported=mirror_unsupported)
    elif input_format == 'tflite' and output_format == 'nnef':
        from .conversion.tflite_to_nnef import Converter
        return Converter(io_transpose=io_transpose,
                         custom_transforms=custom_transforms,
                         custom_functions=custom_functions,
                         mirror_unsupported=mirror_unsupported,
                         keep_io_names=keep_io_names)
    elif input_format == 'nnef' and output_format == 'tflite':
        from .conversion.nnef_to_tflite import Converter
        return Converter(io_transpose=io_transpose,
                         custom_transforms=custom_transforms,
                         custom_functions=custom_functions,
                         mirror_unsupported=mirror_unsupported)
    elif (input_format == 'onnx' or input_format == 'caffe2' or input_format == 'caffe') and output_format == 'nnef':
        from .conversion.onnx_to_nnef import Converter
        return Converter(custom_transforms=custom_transforms,
                         custom_functions=custom_functions,
                         custom_shapes=custom_shapes,
                         infer_shapes=bool(custom_shapes),
                         mirror_unsupported=mirror_unsupported,
                         keep_io_names=keep_io_names,
                         io_transpose=io_transpose)
    elif input_format == 'nnef' and (output_format == 'onnx' or output_format == 'caffe2'):
        from .conversion.nnef_to_onnx import Converter
        return Converter(custom_transforms=custom_transforms,
                         custom_functions=custom_functions,
                         mirror_unsupported=mirror_unsupported)
    else:
        return None


def get_optimizer(format, custom_optimizers=None, dequantize=False):
    if format == 'nnef':
        from .optimization.nnef_optimizer import Optimizer
        return Optimizer(custom_optimizers=custom_optimizers, dequantize=dequantize)
    elif format == 'tf':
        from .optimization.tf_optimizer import Optimizer
        return Optimizer(custom_optimizers=custom_optimizers)
    elif format == 'tflite':
        from .optimization.tflite_optimizer import Optimizer
        return Optimizer(custom_optimizers=custom_optimizers)
    elif format == 'onnx':
        from .optimization.onnx_optimizer import Optimizer
        return Optimizer(custom_optimizers=custom_optimizers)
    else:
        return None


def get_custom_converters(module_names):
    CUSTOM_TRANSFORMS = "CUSTOM_TRANSFORMS"

    transforms = {}
    functions = {}
    for module_name in module_names:
        module = importlib.import_module(module_name)
        if hasattr(module, CUSTOM_TRANSFORMS):
            transforms.update(getattr(module, CUSTOM_TRANSFORMS))

        functions.update(Converter.find_public_functions(module))

    return transforms, functions


def get_custom_shapes(module_names):
    CUSTOM_SHAPES = "CUSTOM_SHAPES"

    shapes = {}
    for module_name in module_names:
        module = importlib.import_module(module_name)
        if hasattr(module, CUSTOM_SHAPES):
            shapes.update(getattr(module, CUSTOM_SHAPES))

    return shapes


def get_custom_fragments(module_names):
    CUSTOM_FRAGMENTS = "CUSTOM_FRAGMENTS"

    fragments = {}
    for module_name in module_names:
        module = importlib.import_module(module_name)
        if hasattr(module, CUSTOM_FRAGMENTS):
            fragments.update(getattr(module, CUSTOM_FRAGMENTS))

    return fragments


def get_custom_optimizers(module_names):
    CUSTOM_OPTIMIZERS = "CUSTOM_OPTIMIZERS"

    optimizers = {}
    for module_name in module_names:
        module = importlib.import_module(module_name)
        if hasattr(module, CUSTOM_OPTIMIZERS):
            optimizers.update(getattr(module, CUSTOM_OPTIMIZERS))

    return optimizers


def needs_conversion(input_format, output_format):
    if input_format == 'caffe2' and output_format == 'onnx':
        return False
    elif input_format == 'onnx' and output_format == 'caffe2':
        return False
    elif input_format == 'caffe' and (output_format == 'onnx' or output_format == 'caffe2'):
        return False
    else:
        return input_format != output_format


def check_nan_or_inf(graph, which):
    valid = True
    for tensor in graph.tensors:
        if tensor.data is not None:
            if np.any(np.isnan(tensor.data)):
                print("{} graph contains nan in tensor '{}'".format(which, tensor.name))
                valid = False
            if np.any(np.isinf(tensor.data)):
                print("{} graph contains inf in tensor '{}'".format(which, tensor.name))
                valid = False

    for op in graph.operations:
        for key, value in six.iteritems(op.attribs):
            if isinstance(value, np.ndarray) and np.issubdtype(value.dtype.type, np.floating):
                if np.any(np.isnan(value)):
                    print("{} graph contains nan in attribute '{}' of operator '{}'".format(which, key, op.type) +
                          " named '{}'".format(op.name) if op.name is not None else "")
                    valid = False
                if np.any(np.isinf(value)):
                    print("{} graph contains inf in attribute '{}' of operator '{}'".format(which, key, op.type) +
                          " named '{}'".format(op.name) if op.name is not None else "")
                    valid = False

    return valid


def main(args):
    io_transpose = False if args.io_transpose is None else True if len(args.io_transpose) == 0 else args.io_transpose

    custom_transforms, custom_functions = get_custom_converters(args.custom_converters) \
        if args.custom_converters is not None else (None, None)

    custom_shapes = get_custom_shapes(args.custom_shapes) or {} if args.custom_shapes is not None else {}

    converter = None
    if needs_conversion(args.input_format, args.output_format):
        converter = get_converter(args.input_format, args.output_format, io_transpose,
                                  custom_transforms, custom_functions, custom_shapes,
                                  args.mirror_unsupported, args.keep_io_names)
        if converter is None:
            print("Unsupported conversion: {} to {}".format(args.input_format, args.output_format))
            return -1

    decomposed = converter.decomposed_operations() if converter else []
    fragments = converter.defined_operations() if converter else {}
    dependencies = converter.defined_operation_dependencies() if converter else {}

    if args.decompose is not None:
        decomposed += args.decompose

    if converter is not None:
        custom_shapes.update(converter.defined_shapes())

    if args.custom_fragments is not None:
        fragments.update(get_custom_fragments(args.custom_fragments))

    reader = get_reader(args.input_format, decomposed=decomposed, fold_constants=args.fold_constants,
                        custom_shapes=custom_shapes)
    if reader is None:
        print("Unsupported input-format: {}".format(args.input_format))
        return -1

    writer = get_writer(args.output_format,
                        fragments=fragments, fragment_dependencies=dependencies,
                        generate_fragments=args.generate_custom_fragments,
                        annotate_shapes=args.annotate_shapes, compression=args.compress)
    if writer is None:
        print("Unsupported output-format: {}".format(args.output_format))
        return -1

    default_output_model = args.input_model + '.' + (args.output_format if args.output_format != 'tf' else 'pb')

    reader_kwargs = {}
    if args.input_shapes is not None:
        input_shapes = eval(args.input_shapes)
        if not isinstance(input_shapes, dict) or not all(isinstance(name, str) and isinstance(shape, tuple)
                                                        for name, shape in six.iteritems(input_shapes)):
            print("'Input-shape' must be a dict of strings to tuples")
            return -1

        reader_kwargs['input_shapes'] = input_shapes

    try:
        graph = reader(args.input_model, **reader_kwargs)

        if not check_nan_or_inf(graph, 'Input'):
            return -1

        if args.input_names is not None or args.output_names is not None:
            not_found_names = []

            if args.input_names is not None:
                input_names = set(args.input_names)
                inputs = [tensor for tensor in graph.tensors if tensor.name in input_names]

                if len(inputs) != len(input_names):
                    found_names = [tensor.name for tensor in inputs]
                    not_found_names.append([input_name for input_name in input_names if input_name not in found_names])
                else:
                    graph.inputs = inputs

            if args.output_names is not None:
                output_names = set(args.output_names)
                outputs = [tensor for tensor in graph.tensors if tensor.name in output_names]

                if len(outputs) != len(output_names):
                    found_names = [tensor.name for tensor in outputs]
                    not_found_names.append([output_name for output_name in output_names if output_name not in found_names])
                else:
                    graph.outputs = outputs

            if len(not_found_names) > 0:
                print("Could not find tensor(s) in graph: {}".format(not_found_names))
                return -1

            utils.remove_unreachable(graph)

        optimizer = get_optimizer(args.input_format)
        if optimizer:
            graph = optimizer(graph, only_required=True)

            if not check_nan_or_inf(graph, 'Optimized input'):
                return -1

        if args.static_only:
            utils.remove_dynamic(graph)
            utils.remove_unreachable(graph)

        if converter:
            graph.sort()
            graph = converter(graph)

            if not check_nan_or_inf(graph, 'Converted'):
                return -1

        tensor_mapping = converter.tensor_mapping() if args.tensor_mapping is not None and converter else None

        if args.optimize:
            custom_optimizers = get_custom_optimizers(args.custom_optimizers) if args.custom_optimizers is not None else None
            optimizer = get_optimizer(args.output_format, custom_optimizers=custom_optimizers, dequantize=args.dequantize)
            if optimizer:
                tensor_lookup = {tensor.name: tensor for tensor in graph.tensors if tensor.name is not None} \
                    if args.tensor_mapping is not None else None

                graph = optimizer(graph)

                if not check_nan_or_inf(graph, 'Optimized output'):
                    return -1

                if args.tensor_mapping is not None:
                    if converter:
                        tensor_mapping = {src: tensor_lookup[dst].name for src, dst in six.iteritems(tensor_mapping)
                                          if tensor_lookup[dst].graph is graph}
                    else:
                        tensor_mapping = {name: tensor.name for name, tensor in six.iteritems(tensor_lookup)
                                          if tensor.graph is graph}

        writer(graph, args.output_model or default_output_model)
        print("Written '{}'".format(args.output_model or default_output_model))

        if args.tensor_mapping is not None:
            with open(args.tensor_mapping, 'w') as file:
                json.dump(tensor_mapping, file, indent=4)

            print("Written '{}'".format(args.tensor_mapping))

        return 0
    except IOError as e:
        print(e)
        return -1
    except ConversionError as e:
        print(e)
        if e.details:
            for detail in e.details:
                print(detail)
        return -1


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--input-model', type=str, required=True,
                        help='The input model')
    parser.add_argument('--output-model', type=str, default=None,
                        help='The output model')
    parser.add_argument('--input-format', type=str, required=True,
                        choices=['tf', 'tflite', 'onnx', 'nnef', 'caffe2', 'caffe'],
                        help='The format of the input model')
    parser.add_argument('--output-format', type=str, required=True,
                        choices=['tf', 'tflite', 'onnx', 'nnef', 'caffe2'],
                        help='The format of the output model')
    parser.add_argument('--input-shapes', type=str, default=None,
                        help='The (dict of) shape(s) to use for input(s).')
    parser.add_argument('--io-transpose', type=str, nargs='*', default=None,
                        help='The inputs/outputs to transpose')
    parser.add_argument('--fold-constants', action='store_true',
                        help='Enable folding of constant ops')
    parser.add_argument('--optimize', action='store_true',
                        help='Turn on optimization of resulting NNEF model')
    parser.add_argument('--dequantize', action='store_true',
                        help='Dequantize the weights of a quantized network and omit quantization parameters')
    parser.add_argument('--custom-converters', type=str, nargs='+',
                        help='Module(s) containing custom converter code')
    parser.add_argument('--custom-shapes', type=str, nargs='+',
                        help='Module(s) containing custom shape inference code (when converting to/from NNEF)')
    parser.add_argument('--custom-fragments', type=str, nargs='+',
                        help='Module(s) containing custom fragment code (when converting to NNEF)')
    parser.add_argument('--custom-optimizers', type=str, nargs='+',
                        help='Module(s) containing custom optimizer code (when converting to NNEF)')
    parser.add_argument('--mirror-unsupported', action='store_true',
                        help='Enable mirror-conversion of unsupported operations')
    parser.add_argument('--generate-custom-fragments', action='store_true',
                        help='Enable automatic generation of fragments for custom operations')
    parser.add_argument('--keep-io-names', action='store_true',
                        help='Keep the names of model inputs/outputs if possible')
    parser.add_argument('--decompose', type=str, nargs='*', default=None,
                        help='Names of operators to be decomposed by NNEF parser')
    parser.add_argument('--input-names', type=str, nargs='+',
                        help='Names of input tensor where the graph is cut before conversion')
    parser.add_argument('--output-names', type=str, nargs='+',
                        help='Names of output tensor where the graph is cut before conversion')
    parser.add_argument('--static-only', action='store_true',
                        help='Only convert static part of the graph, for which tensor shapes are known')
    parser.add_argument('--tensor-mapping', type=str, nargs='?', default=None, const='tensor_mapping.json',
                        help='Export mapping of tensor names from input to output model')
    parser.add_argument('--annotate-shapes', action='store_true',
                        help='Add tensor shapes as comments to NNEF output model')
    parser.add_argument('--compress', type=int, nargs='?', default=None, const=1,
                        help='Compress output NNEF folder at the given compression level')
    exit(main(parser.parse_args()))



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class SparsityParameters(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsSparsityParameters(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SparsityParameters()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def SparsityParametersBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # SparsityParameters
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SparsityParameters
    def TraversalOrder(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # SparsityParameters
    def TraversalOrderAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # SparsityParameters
    def TraversalOrderLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SparsityParameters
    def TraversalOrderIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # SparsityParameters
    def BlockMap(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # SparsityParameters
    def BlockMapAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # SparsityParameters
    def BlockMapLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SparsityParameters
    def BlockMapIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # SparsityParameters
    def DimMetadata(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .DimensionMetadata import DimensionMetadata
            obj = DimensionMetadata()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SparsityParameters
    def DimMetadataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SparsityParameters
    def DimMetadataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def SparsityParametersStart(builder): builder.StartObject(3)
def SparsityParametersAddTraversalOrder(builder, traversalOrder): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(traversalOrder), 0)
def SparsityParametersStartTraversalOrderVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def SparsityParametersAddBlockMap(builder, blockMap): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(blockMap), 0)
def SparsityParametersStartBlockMapVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def SparsityParametersAddDimMetadata(builder, dimMetadata): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(dimMetadata), 0)
def SparsityParametersStartDimMetadataVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def SparsityParametersEnd(builder): return builder.EndObject()



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import flatbuffers as fb
import numpy as np
import sys
import re

# See: https://github.com/tensorflow/tensorflow/blob/master/tensorflow/lite/schema/schema.fbs


OUTPUT_FILE_IDENTIFIER = "TFL3"
OUTPUT_SCHEMA_VERSION = 3

BuiltinOptionsClasses = [
    None,
    fb.Conv2DOptions,
    fb.DepthwiseConv2DOptions,
    fb.ConcatEmbeddingsOptions,
    fb.LSHProjectionOptions,
    fb.Pool2DOptions,
    fb.SVDFOptions,
    fb.RNNOptions,
    fb.FullyConnectedOptions,
    fb.SoftmaxOptions,
    fb.ConcatenationOptions,
    fb.AddOptions,
    fb.L2NormOptions,
    fb.LocalResponseNormalizationOptions,
    fb.LSTMOptions,
    fb.ResizeBilinearOptions,
    fb.CallOptions,
    fb.ReshapeOptions,
    fb.SkipGramOptions,
    fb.SpaceToDepthOptions,
    fb.EmbeddingLookupSparseOptions,
    fb.MulOptions,
    fb.PadOptions,
    fb.GatherOptions,
    fb.BatchToSpaceNDOptions,
    fb.SpaceToBatchNDOptions,
    fb.TransposeOptions,
    fb.ReducerOptions,
    fb.SubOptions,
    fb.DivOptions,
    fb.SqueezeOptions,
    fb.SequenceRNNOptions,
    fb.StridedSliceOptions,
    fb.ExpOptions,
    fb.TopKV2Options,
    fb.SplitOptions,
    fb.LogSoftmaxOptions,
    fb.CastOptions,
    fb.DequantizeOptions,
    fb.MaximumMinimumOptions,
    fb.ArgMaxOptions,
    fb.LessOptions,
    fb.NegOptions,
    fb.PadV2Options,
    fb.GreaterOptions,
    fb.GreaterEqualOptions,
    fb.LessEqualOptions,
    fb.SelectOptions,
    fb.SliceOptions,
    fb.TransposeConvOptions,
    fb.SparseToDenseOptions,
    fb.TileOptions,
    fb.ExpandDimsOptions,
    fb.EqualOptions,
    fb.NotEqualOptions,
    fb.ShapeOptions,
    fb.PowOptions,
    fb.ArgMinOptions,
    fb.FakeQuantOptions,
    fb.PackOptions,
    fb.LogicalOrOptions,
    fb.OneHotOptions,
    fb.LogicalAndOptions,
    fb.LogicalNotOptions,
    fb.UnpackOptions,
    fb.FloorDivOptions,
    fb.SquareOptions,
    fb.ZerosLikeOptions,
    fb.FillOptions,
    fb.BidirectionalSequenceLSTMOptions,
    fb.BidirectionalSequenceRNNOptions,
    fb.UnidirectionalSequenceLSTMOptions,
    fb.FloorModOptions,
    fb.RangeOptions,
    fb.ResizeNearestNeighborOptions,
    fb.LeakyReluOptions,
    fb.SquaredDifferenceOptions,
    fb.MirrorPadOptions,
    fb.AbsOptions,
    fb.SplitVOptions,
    fb.UniqueOptions,
    fb.ReverseV2Options,
    fb.AddNOptions,
    fb.GatherNdOptions,
    fb.CosOptions,
    fb.WhereOptions,
    fb.RankOptions,
    fb.ReverseSequenceOptions,
    fb.MatrixDiagOptions,
    fb.QuantizeOptions,
    fb.MatrixSetDiagOptions,
    fb.HardSwishOptions,
    fb.IfOptions,
    fb.WhileOptions,
    fb.DepthToSpaceOptions,
    fb.NonMaxSuppressionV4Options,
    fb.NonMaxSuppressionV5Options,
    fb.ScatterNdOptions,
    fb.SelectV2Options,
    fb.DensifyOptions,
    fb.SegmentSumOptions,
    fb.BatchMatMulOptions,
]

BuiltinOptionsByOperator = {
    fb.BuiltinOperator.ADD: fb.BuiltinOptions.AddOptions,
    fb.BuiltinOperator.AVERAGE_POOL_2D: fb.BuiltinOptions.Pool2DOptions,
    fb.BuiltinOperator.CONCATENATION: fb.BuiltinOptions.ConcatenationOptions,
    fb.BuiltinOperator.CONV_2D: fb.BuiltinOptions.Conv2DOptions,
    fb.BuiltinOperator.DEPTHWISE_CONV_2D: fb.BuiltinOptions.DepthwiseConv2DOptions,
    fb.BuiltinOperator.DEQUANTIZE: fb.BuiltinOptions.DequantizeOptions,
    fb.BuiltinOperator.EMBEDDING_LOOKUP: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.FLOOR: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.FULLY_CONNECTED: fb.BuiltinOptions.FullyConnectedOptions,
    fb.BuiltinOperator.HASHTABLE_LOOKUP: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.L2_NORMALIZATION: fb.BuiltinOptions.L2NormOptions,
    fb.BuiltinOperator.L2_POOL_2D: fb.BuiltinOptions.Pool2DOptions,
    fb.BuiltinOperator.LOCAL_RESPONSE_NORMALIZATION: fb.BuiltinOptions.LocalResponseNormalizationOptions,
    fb.BuiltinOperator.LOGISTIC: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.LSH_PROJECTION: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.LSTM: fb.BuiltinOptions.LSTMOptions,
    fb.BuiltinOperator.MAX_POOL_2D: fb.BuiltinOptions.Pool2DOptions,
    fb.BuiltinOperator.MUL: fb.BuiltinOptions.MulOptions,
    fb.BuiltinOperator.RELU: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.RELU_N1_TO_1: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.RELU6: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.RESHAPE: fb.BuiltinOptions.ReshapeOptions,
    fb.BuiltinOperator.RESIZE_BILINEAR: fb.BuiltinOptions.ResizeBilinearOptions,
    fb.BuiltinOperator.RNN: fb.BuiltinOptions.RNNOptions,
    fb.BuiltinOperator.SOFTMAX: fb.BuiltinOptions.SoftmaxOptions,
    fb.BuiltinOperator.SPACE_TO_DEPTH: fb.BuiltinOptions.SpaceToDepthOptions,
    fb.BuiltinOperator.SVDF: fb.BuiltinOptions.SVDFOptions,
    fb.BuiltinOperator.TANH: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.CONCAT_EMBEDDINGS: fb.BuiltinOptions.ConcatEmbeddingsOptions,
    fb.BuiltinOperator.SKIP_GRAM: fb.BuiltinOptions.SkipGramOptions,
    fb.BuiltinOperator.CALL: fb.BuiltinOptions.CallOptions,
    fb.BuiltinOperator.CUSTOM: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.EMBEDDING_LOOKUP_SPARSE: fb.BuiltinOptions.EmbeddingLookupSparseOptions,
    fb.BuiltinOperator.PAD: fb.BuiltinOptions.PadOptions,
    fb.BuiltinOperator.UNIDIRECTIONAL_SEQUENCE_RNN: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.GATHER: fb.BuiltinOptions.GatherOptions,
    fb.BuiltinOperator.BATCH_TO_SPACE_ND: fb.BuiltinOptions.BatchToSpaceNDOptions,
    fb.BuiltinOperator.SPACE_TO_BATCH_ND: fb.BuiltinOptions.SpaceToBatchNDOptions,
    fb.BuiltinOperator.TRANSPOSE: fb.BuiltinOptions.TransposeOptions,
    fb.BuiltinOperator.MEAN: fb.BuiltinOptions.ReducerOptions,
    fb.BuiltinOperator.SUB: fb.BuiltinOptions.SubOptions,
    fb.BuiltinOperator.DIV: fb.BuiltinOptions.DivOptions,
    fb.BuiltinOperator.SQUEEZE: fb.BuiltinOptions.SqueezeOptions,
    fb.BuiltinOperator.UNIDIRECTIONAL_SEQUENCE_LSTM: fb.BuiltinOptions.UnidirectionalSequenceLSTMOptions,
    fb.BuiltinOperator.STRIDED_SLICE: fb.BuiltinOptions.StridedSliceOptions,
    fb.BuiltinOperator.BIDIRECTIONAL_SEQUENCE_RNN: fb.BuiltinOptions.BidirectionalSequenceRNNOptions,
    fb.BuiltinOperator.EXP: fb.BuiltinOptions.ExpOptions,
    fb.BuiltinOperator.TOPK_V2: fb.BuiltinOptions.TopKV2Options,
    fb.BuiltinOperator.SPLIT: fb.BuiltinOptions.SplitOptions,
    fb.BuiltinOperator.LOG_SOFTMAX: fb.BuiltinOptions.LogSoftmaxOptions,
    fb.BuiltinOperator.DELEGATE: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.BIDIRECTIONAL_SEQUENCE_LSTM: fb.BuiltinOptions.BidirectionalSequenceLSTMOptions,
    fb.BuiltinOperator.CAST: fb.BuiltinOptions.CastOptions,
    fb.BuiltinOperator.PRELU: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.MAXIMUM: fb.BuiltinOptions.MaximumMinimumOptions,
    fb.BuiltinOperator.ARG_MAX: fb.BuiltinOptions.ArgMaxOptions,
    fb.BuiltinOperator.MINIMUM: fb.BuiltinOptions.MaximumMinimumOptions,
    fb.BuiltinOperator.LESS: fb.BuiltinOptions.LessOptions,
    fb.BuiltinOperator.NEG: fb.BuiltinOptions.NegOptions,
    fb.BuiltinOperator.PADV2: fb.BuiltinOptions.PadV2Options,
    fb.BuiltinOperator.GREATER: fb.BuiltinOptions.GreaterOptions,
    fb.BuiltinOperator.GREATER_EQUAL: fb.BuiltinOptions.GreaterEqualOptions,
    fb.BuiltinOperator.LESS_EQUAL: fb.BuiltinOptions.LessEqualOptions,
    fb.BuiltinOperator.SELECT: fb.BuiltinOptions.SelectOptions,
    fb.BuiltinOperator.SLICE: fb.BuiltinOptions.SliceOptions,
    fb.BuiltinOperator.SIN: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.TRANSPOSE_CONV: fb.BuiltinOptions.TransposeConvOptions,
    fb.BuiltinOperator.SPARSE_TO_DENSE: fb.BuiltinOptions.SparseToDenseOptions,
    fb.BuiltinOperator.TILE: fb.BuiltinOptions.TileOptions,
    fb.BuiltinOperator.EXPAND_DIMS: fb.BuiltinOptions.ExpandDimsOptions,
    fb.BuiltinOperator.EQUAL: fb.BuiltinOptions.EqualOptions,
    fb.BuiltinOperator.NOT_EQUAL: fb.BuiltinOptions.NotEqualOptions,
    fb.BuiltinOperator.LOG: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.SUM: fb.BuiltinOptions.ReducerOptions,
    fb.BuiltinOperator.SQRT: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.RSQRT: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.SHAPE: fb.BuiltinOptions.ShapeOptions,
    fb.BuiltinOperator.POW: fb.BuiltinOptions.PowOptions,
    fb.BuiltinOperator.ARG_MIN: fb.BuiltinOptions.ArgMinOptions,
    fb.BuiltinOperator.FAKE_QUANT: fb.BuiltinOptions.FakeQuantOptions,
    fb.BuiltinOperator.REDUCE_PROD: fb.BuiltinOptions.ReducerOptions,
    fb.BuiltinOperator.REDUCE_MAX: fb.BuiltinOptions.ReducerOptions,
    fb.BuiltinOperator.PACK: fb.BuiltinOptions.PackOptions,
    fb.BuiltinOperator.LOGICAL_OR: fb.BuiltinOptions.LogicalOrOptions,
    fb.BuiltinOperator.ONE_HOT: fb.BuiltinOptions.OneHotOptions,
    fb.BuiltinOperator.LOGICAL_AND: fb.BuiltinOptions.LogicalAndOptions,
    fb.BuiltinOperator.LOGICAL_NOT: fb.BuiltinOptions.LogicalNotOptions,
    fb.BuiltinOperator.UNPACK: fb.BuiltinOptions.UnpackOptions,
    fb.BuiltinOperator.REDUCE_MIN: fb.BuiltinOptions.ReducerOptions,
    fb.BuiltinOperator.FLOOR_DIV: fb.BuiltinOptions.FloorDivOptions,
    fb.BuiltinOperator.REDUCE_ANY: fb.BuiltinOptions.ReducerOptions,
    fb.BuiltinOperator.SQUARE: fb.BuiltinOptions.SquareOptions,
    fb.BuiltinOperator.ZEROS_LIKE: fb.BuiltinOptions.ZerosLikeOptions,
    fb.BuiltinOperator.FILL: fb.BuiltinOptions.FillOptions,
    fb.BuiltinOperator.FLOOR_MOD: fb.BuiltinOptions.FloorModOptions,
    fb.BuiltinOperator.RANGE: fb.BuiltinOptions.RangeOptions,
    fb.BuiltinOperator.RESIZE_NEAREST_NEIGHBOR: fb.BuiltinOptions.ResizeNearestNeighborOptions,
    fb.BuiltinOperator.LEAKY_RELU: fb.BuiltinOptions.LeakyReluOptions,
    fb.BuiltinOperator.SQUARED_DIFFERENCE: fb.BuiltinOptions.SquaredDifferenceOptions,
    fb.BuiltinOperator.MIRROR_PAD: fb.BuiltinOptions.MirrorPadOptions,
    fb.BuiltinOperator.ABS: fb.BuiltinOptions.AbsOptions,
    fb.BuiltinOperator.SPLIT_V: fb.BuiltinOptions.SplitVOptions,
    fb.BuiltinOperator.UNIQUE: fb.BuiltinOptions.UniqueOptions,
    fb.BuiltinOperator.CEIL: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.REVERSE_V2: fb.BuiltinOptions.ReverseV2Options,
    fb.BuiltinOperator.ADD_N: fb.BuiltinOptions.AddNOptions,
    fb.BuiltinOperator.GATHER_ND: fb.BuiltinOptions.GatherNdOptions,
    fb.BuiltinOperator.COS: fb.BuiltinOptions.CosOptions,
    fb.BuiltinOperator.WHERE: fb.BuiltinOptions.WhereOptions,
    fb.BuiltinOperator.RANK: fb.BuiltinOptions.RankOptions,
    fb.BuiltinOperator.ELU: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.REVERSE_SEQUENCE: fb.BuiltinOptions.ReverseSequenceOptions,
    fb.BuiltinOperator.MATRIX_DIAG: fb.BuiltinOptions.MatrixDiagOptions,
    fb.BuiltinOperator.QUANTIZE: fb.BuiltinOptions.QuantizeOptions,
    fb.BuiltinOperator.MATRIX_SET_DIAG: fb.BuiltinOptions.MatrixSetDiagOptions,
    fb.BuiltinOperator.HARD_SWISH: fb.HardSwishOptions,
    fb.BuiltinOperator.IF: fb.IfOptions,
    fb.BuiltinOperator.WHILE: fb.WhileOptions,
    fb.BuiltinOperator.DEPTH_TO_SPACE: fb.DepthToSpaceOptions,
    fb.BuiltinOperator.NON_MAX_SUPPRESSION_V4: fb.NonMaxSuppressionV4Options,
    fb.BuiltinOperator.NON_MAX_SUPPRESSION_V5: fb.NonMaxSuppressionV5Options,
    fb.BuiltinOperator.SCATTER_ND: fb.ScatterNdOptions,
    fb.BuiltinOperator.ROUND: fb.BuiltinOptions.NONE,
    fb.BuiltinOperator.SELECT_V2: fb.BuiltinOptions.SelectV2Options,
    fb.BuiltinOperator.DENSIFY: fb.BuiltinOptions.DensifyOptions,
    fb.BuiltinOperator.SEGMENT_SUM: fb.BuiltinOptions.SegmentSumOptions,
    fb.BuiltinOperator.BATCH_MATMUL: fb.BuiltinOptions.BatchMatMulOptions,
}

CustomOptionsKey = 'custom_options'

DtypeToNumpy = {
    fb.TensorType.FLOAT16: np.float16,
    fb.TensorType.FLOAT32: np.float32,
    fb.TensorType.INT8: np.int8,
    fb.TensorType.INT16: np.int16,
    fb.TensorType.INT32: np.int32,
    fb.TensorType.INT64: np.int64,
    fb.TensorType.UINT8: np.uint8,
    fb.TensorType.STRING: np.str_,
    fb.TensorType.BOOL: np.bool_,
    fb.TensorType.COMPLEX64: np.complex64,
}

DtypeFromNumpy = {
    np.float16: fb.TensorType.FLOAT16,
    np.float32: fb.TensorType.FLOAT32,
    np.int8: fb.TensorType.INT8,
    np.int16: fb.TensorType.INT16,
    np.int32: fb.TensorType.INT32,
    np.int64: fb.TensorType.INT64,
    np.uint8: fb.TensorType.UINT8,
    np.str_: fb.TensorType.STRING,
    np.bool_: fb.TensorType.BOOL,
    np.complex64: fb.TensorType.COMPLEX64,
}


_regex1 = re.compile('(.)([A-Z][a-z]+)')
_regex2 = re.compile('([a-z0-9])([A-Z])')


def camel_to_snake(s):
    subbed = _regex1.sub(r'\1_\2', s)
    return _regex2.sub(r'\1_\2', subbed).lower()


def snake_to_camel(s):
    return ''.join(c for c in s.title() if c != '_')


def substitute_enum_value_with_name(key, value, optionsClass):
    cls, map = _OptionEnumNameByValueMaps.get(key, (None, None))
    return map[value] if map is not None and (cls is None or cls == optionsClass) else value


def substitute_enum_name_with_value(key, name, optionsClass):
    cls, map = _OptionEnumValueByNameMaps.get(key, (None, None))
    return map[name] if map is not None and (cls is None or cls == optionsClass) else name


def _generate_enum_value_by_name(enumClass):
    return {name: value for name, value in enumClass.__dict__.items() if not name.startswith('_')}


def _generate_enum_name_by_value(enumClass):
    return {value: name for name, value in enumClass.__dict__.items() if not name.startswith('_')}


def enumerate_options_getters(optionsClass):
    return {camel_to_snake(name): func for name, func in optionsClass.__dict__.items()
            if not name.startswith('_')
            and name != 'Init' and not name.startswith('GetRootAs')
            and not name.endswith('AsNumpy') and not name.endswith('Length')
            and not isinstance(func, classmethod)}


def enumerate_options_length_getters(optionsClass):
    return {camel_to_snake(name[:-6]): func for name, func in optionsClass.__dict__.items()
            if not name.startswith('_')
            and not name.startswith('GetRootAs') and name.endswith('Length')}


def enumerate_options_adders(optionsClass):
    className = optionsClass.__name__
    prefix = className + 'Add'
    optionsModule = sys.modules[optionsClass.__module__]
    return {camel_to_snake(name[len(prefix):]): func for name, func in optionsModule.__dict__.items()
            if name.startswith(prefix)}


def enumerate_options_vector_starters(optionsClass):
    className = optionsClass.__name__
    prefix, suffix = className + 'Start', 'Vector'
    optionsModule = sys.modules[optionsClass.__module__]
    return {camel_to_snake(name[len(prefix):-len(suffix)]): func for name, func in optionsModule.__dict__.items()
            if name.startswith(prefix) and name.endswith(suffix)}


def get_options_starter_ender(optionsClass):
    className = optionsClass.__name__
    optionsModule = sys.modules[optionsClass.__module__]
    moduleDict = optionsModule.__dict__
    return moduleDict[className + 'Start'], moduleDict[className + 'End']


_OptionEnumNameByValueMaps = {
    'padding': (None, _generate_enum_name_by_value(fb.Padding)),
    'fused_activation_function': (None, _generate_enum_name_by_value(fb.ActivationFunctionType)),
    'weights_format': (
        fb.FullyConnectedOptions, _generate_enum_name_by_value(fb.FullyConnectedOptionsWeightsFormat)),
    'type': (fb.LSHProjectionOptions, _generate_enum_name_by_value(fb.LSHProjectionType)),
    'kernel_type': (fb.LSTMOptions, _generate_enum_name_by_value(fb.LSTMKernelType)),
    'combiner': (fb.EmbeddingLookupSparseOptions, _generate_enum_name_by_value(fb.CombinerType)),
}

_OptionEnumValueByNameMaps = {
    'padding': (None, _generate_enum_value_by_name(fb.Padding)),
    'fused_activation_function': (None, _generate_enum_value_by_name(fb.ActivationFunctionType)),
    'weights_format': (
        fb.FullyConnectedOptions, _generate_enum_value_by_name(fb.FullyConnectedOptionsWeightsFormat)),
    'type': (fb.LSHProjectionOptions, _generate_enum_value_by_name(fb.LSHProjectionType)),
    'kernel_type': (fb.LSTMOptions, _generate_enum_value_by_name(fb.LSTMKernelType)),
    'combiner': (fb.EmbeddingLookupSparseOptions, _generate_enum_value_by_name(fb.CombinerType)),
}

BuiltinOperatorTypeByValue = _generate_enum_name_by_value(fb.BuiltinOperator)
BuiltinOperatorValueByType = _generate_enum_value_by_name(fb.BuiltinOperator)



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable


def _split_counter_from_name(str):
    if len(str) > 0 and not str[-1].isdigit():
        return str, None

    i = len(str)
    while i > 0:
        if not str[i-1].isdigit():
            return str[:i], int(str[i:])
        i -= 1
    return None, int(str)


def generate_tensor_names_from_op_type(graph, keep_io_names=False):
    used_names = set()
    if keep_io_names:
        used_names.update(tensor.name for tensor in graph.inputs if tensor.name is not None)
        used_names.update(tensor.name for tensor in graph.outputs if tensor.name is not None)

    op_counts = {}

    for op in graph.operations:
        for tensor in op.outputs:
            if keep_io_names and tensor.name is not None and (tensor in graph.inputs or tensor in graph.outputs):
                continue

            idx = op_counts.get(op.type, 0) + 1
            while op.type + str(idx) in used_names:
                idx += 1

            op_counts[op.type] = idx
            tensor.name = op.type + str(idx)

    for tensor in graph.tensors:
        if tensor.producer is None:
            tensor.name = None


def generate_missing_tensor_names_from_op_type(graph):
    counters = {}
    for tensor in graph.tensors:
        if tensor.name is not None:
            name, count = _split_counter_from_name(tensor.name)
            if name is not None and count is not None:
                counters[name] = max(counters.get(name, 0), count)

    for tensor in graph.tensors:
        if tensor.name is None and tensor.producer is not None:
            op = tensor.producer
            idx = counters.get(op.type, 0) + 1
            counters[op.type] = idx
            tensor.name = op.type + str(idx)


def generate_op_names_from_op_type(graph):
    op_counts = {}
    for op in graph.operations:
        idx = op_counts.get(op.type, 0) + 1
        op_counts[op.type] = idx
        op.name = op.type + str(idx)


def replace_tensor_in_graph_inputs(graph, old_tensor, new_tensor):
    graph.inputs = [new_tensor if t is old_tensor else t for t in graph.inputs]


def replace_tensor_in_graph_outputs(graph, old_tensor, new_tensor):
    graph.outputs = [new_tensor if t is old_tensor else t for t in graph.outputs]


def replace_tensor_in_consumers(graph, old_tensor, new_tensor):
    for consumer in list(old_tensor.consumers):     # copy list to avoid changes during iteration
        sequence = tuple if isinstance(consumer.inputs, tuple) else list
        consumer.inputs = sequence(new_tensor if t is old_tensor else t for t in consumer.inputs)

    replace_tensor_in_graph_outputs(graph, old_tensor, new_tensor)


def replace_tensor_in_producers(graph, old_tensor, new_tensor):
    for producer in list(old_tensor.producers):     # copy list to avoid changes during iteration
        sequence = tuple if isinstance(producer.outputs, tuple) else list
        producer.outputs = sequence(new_tensor if t is old_tensor else t for t in producer.outputs)

    replace_tensor_in_graph_inputs(graph, old_tensor, new_tensor)


def bypass_and_remove(graph, op, remove_input_not_output=False):
    assert len(op.outputs) == 1 and len(op.inputs) == 1

    op_input = op.input
    op_output = op.output

    graph.remove_operation(op, unlink=True)

    if remove_input_not_output:
        replace_tensor_in_consumers(graph, op_input, op_output)
        replace_tensor_in_producers(graph, op_input, op_output)
        graph.remove_tensor(op_input)
    else:
        replace_tensor_in_consumers(graph, op_output, op_input)
        replace_tensor_in_producers(graph, op_output, op_input)
        graph.remove_tensor(op_output)


def replace_chain(graph, types, func, allow_forks=False):
    def _match_type(type, template):
        return type == template if isinstance(template, str) else\
            type in template if isinstance(template, Iterable) else False

    def _match_link(op, template, is_last):
        return _match_type(op.type, template) and (len(op.outputs) == 1 or is_last)

    def _match_chain(op, types, allow_forks):
        if not _match_link(op, types[0], is_last=len(types) == 1):
            return None

        chain = [op]
        tensor = op.output
        for idx, type in enumerate(types[1:]):
            is_last = idx + 1 == len(types) - 1

            if not allow_forks and len(tensor.consumers) > 1:
                return None

            op = next((consumer for consumer in tensor.consumers if _match_link(consumer, type, is_last)), None)
            if op is None:
                return None

            chain.append(op)
            if not is_last:
                tensor = op.output

        return chain

    changed = False
    i = 0
    while i < len(graph.operations):
        count = len(graph.operations)
        chain = _match_chain(graph.operations[i], types, allow_forks)
        if chain is not None and func(*chain) is not False:
            k = i
            while graph.operations[k] is not chain[-1]:
                k += 1

            for j in range(count, len(graph.operations)):
                graph.move_operation(j, k)
                k += 1

            offs = len(chain) - 1
            while offs > 0 and len(chain[offs - 1].output.consumers) == 1:
                offs -= 1

            interns = [op.output for op in chain[offs:-1]]
            graph.remove_operations(chain[offs:], unlink=True)
            graph.remove_tensors(interns)
            changed = True
        else:
            i += 1
    return changed


def remove_unreachable(graph):
    visited = {tensor.producer for tensor in graph.outputs}
    queue = list(visited)

    k = 0
    while k < len(queue):
        op = queue[k]
        k += 1

        for tensor in op.inputs:
            if tensor.producer is not None and tensor.producer not in visited and \
                    (tensor not in graph.inputs or len(tensor.producer.inputs) == 0):
                visited.add(tensor.producer)
                queue.append(tensor.producer)

    graph.remove_operations({op for op in graph.operations if op not in visited}, unlink=True)
    graph.remove_tensors({tensor for tensor in graph.tensors
                          if len(tensor.producers) == 0 and len(tensor.consumers) == 0
                          and tensor not in graph.inputs and tensor not in graph.outputs})


def remove_dynamic(graph):
    dynamic_tensors = {tensor for tensor in graph.tensors if tensor.shape is None or any(s is None for s in tensor.shape)}
    dynamic_ops = {tensor.producer for tensor in dynamic_tensors}

    queue = list(dynamic_ops)

    k = 0
    while k < len(queue):
        op = queue[k]
        k += 1

        for tensor in op.outputs:
            dynamic_tensors.add(tensor)
            for op in tensor.consumers:
                if op not in dynamic_ops:
                    dynamic_ops.add(op)
                    queue.append(op)

    kept_outputs = [tensor for tensor in graph.outputs if tensor not in dynamic_tensors]
    new_outputs = kept_outputs + [tensor for tensor in graph.tensors
                                  if all(op in dynamic_ops for op in tensor.consumers) and tensor not in dynamic_tensors]

    graph.outputs = kept_outputs
    graph.remove_operations(dynamic_ops, unlink=True)
    graph.remove_tensors(dynamic_tensors)
    graph.outputs = new_outputs



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import
from .converter import ConverterToNNEF as _Converter, Transform, ConversionError
from ..model.utils import generate_tensor_names_from_op_type
from ..utils import types
from collections import OrderedDict
import numpy as np


_RELU6_FRAGMENT = """
fragment relu6( input: tensor<scalar> ) -> ( output: tensor<scalar> )
{
    output = clamp(input, 0.0, 6.0);
}
"""

_INT_MAX = 2 ** 31 - 1


class Converter(_Converter):

    _ConvOpTypes = ['Conv1D', 'Conv2D', 'Conv3D', 'Conv2DBackpropInput', 'Conv1DBackpropInput', 'Conv3DBackpropInput']
    _DepthwiseConvOpTypes = ['DepthwiseConv2dNative', 'DepthwiseConv2dNativeBackpropInput']

    @staticmethod
    def defined_operations():
        return {
            'relu6': _RELU6_FRAGMENT,
        }

    def __init__(self, io_transpose=False, custom_transforms=None, custom_functions=None,
                 mirror_unsupported=False, keep_io_names=False):
        _Converter.__init__(self, transforms=self.merge_transforms(_Transforms, custom_transforms),
                            functions=custom_functions, mirror_unsupported=mirror_unsupported)
        self._io_transpose = io_transpose
        self._keep_io_names = keep_io_names

    def __call__(self, graph):
        graph = _Converter.__call__(self, graph)
        self.remove_unused_constants(graph)
        self.inline_scalar_constants(graph)
        self.convert_constants_to_variables(graph)
        self._fix_output_transposes(graph)
        self._ensure_valid_ids(graph)
        generate_tensor_names_from_op_type(graph, keep_io_names=self._keep_io_names)
        return graph

    def _global_attribs(self):
        return {'_lite_': False}

    def _fix_output_transposes(self, graph):
        outputs = [self.transpose_input(tensor) if self.needs_io_transpose(tensor) else
                   self.undo_transpose(tensor) for tensor in graph.outputs]

        if self._keep_io_names:
            for i in range(len(outputs)):
                if outputs[i] is not graph.outputs[i]:
                    outputs[i].name = graph.outputs[i].name

        graph.outputs = outputs

    def _is_conv_filter(self, tensor):
        tensor = self._tensor_map.get(tensor)
        return tensor and len(tensor.consumers) > 0 and \
               all(op.type in Converter._ConvOpTypes and op.inputs[1] is tensor for op in tensor.consumers)

    def _is_depthwise_conv_filter(self, tensor):
        tensor = self._tensor_map.get(tensor)
        return tensor and len(tensor.consumers) > 0 and \
               all(op.type in Converter._DepthwiseConvOpTypes and op.inputs[1] is tensor for op in tensor.consumers)

    def _is_constant(self, tensor):
        if tensor.producer:
            return tensor.producer.type == 'Const'
        else:
            return tensor.data is not None

    def _read_constant(self, tensor, type=None):
        if tensor.producer is None:
            return types.from_numpy(tensor.data, type=type)
        elif tensor.producer.type == 'Const':
            value = tensor.producer.attribs['value']
            return types.from_numpy(value, type=type) if isinstance(value, np.ndarray) else types.cast(value, type=type)
        else:
            raise ConversionError('trying to evaluate non-constant tensor')

    def needs_io_transpose(self, tensor):
        if tensor.rank <= 2:
            return False
        if isinstance(self._io_transpose, bool):
            return self._io_transpose
        else:
            return tensor.name in self._io_transpose

    def is_nxc(self, format):
        return format[0] == 'N' and format[-1] == 'C' and len(format) > 2

    def is_cxn(self, format):
        return format[0] == 'C' and format[-1] == 'N' and len(format) > 2

    def is_xcn(self, format):
        return format[-2] == 'C' and format[-1] == 'N' and len(format) > 2

    def transpose_input(self, tensor, format='NXC'):
        if self.is_nxc(format):
            return self._pre_transpose(tensor, self.nxc_to_ncx_perm(tensor.rank)) \
                if not self.transposing(tensor) and tensor.rank > 2 else tensor
        else:
            assert not self.transposing(tensor)
            return tensor

    def transpose_output(self, tensor, format='NXC'):
        if self.is_nxc(format):
            self._transposes[tensor] = self.nxc_to_ncx(tensor.shape)
        return tensor

    def transpose_filter(self, tensor, format='XCN'):
        if self.is_xcn(format):
            perm = self.xcn_to_ncx_perm(tensor.rank)
        elif self.is_nxc(format):
            perm = self.nxc_to_ncx_perm(tensor.rank)
        elif self.is_cxn(format):
            perm = self.cxn_to_ncx_perm(tensor.rank)
        else:
            assert False

        return self._pre_transpose(tensor, perm)

    def transpose_depthwise_filter(self, tensor, format='XCN'):
        if self.is_xcn(format):
            perm = self.xcn_to_ncx_perm(tensor.rank)
        elif self.is_nxc(format):
            perm = self.nxc_to_ncx_perm(tensor.rank)
        elif self.is_cxn(format):
            perm = self.cxn_to_ncx_perm(tensor.rank)
        else:
            assert False

        shape = tensor.shape[:-2] + (1, -1)
        return self._pre_transpose(self._reshape(tensor, shape), perm)

    def transpose_like(self, tensor, ref):
        if ref is not None and self.transposing(ref):
            self.transpose_output(tensor)
        return tensor

    def undo_transpose(self, tensor):
        perm = self.ncx_to_nxc_perm(tensor.rank)
        if perm == list(range(tensor.rank)):
            return tensor
        return self._pre_transpose(tensor, perm) if self.transposing(tensor) else tensor

    def convert_size(self, value, format):
        return value[1:-1] if self.is_nxc(format) else value[2:]

    def convert_padding(self, padding, rank, explicit_paddings=None, format=None):
        padding = padding.upper()
        if padding == 'SAME':
            return []
        elif padding == 'VALID':
            return [(0, 0)] * rank
        elif padding == 'EXPLICIT':
            assert explicit_paddings is not None and format is not None
            explicit_paddings = list(zip(explicit_paddings[0::2], explicit_paddings[1::2]))
            return explicit_paddings[1:-1] if self.is_nxc(format) else explicit_paddings[2:]
        else:
            assert False, "unknown padding type '{}'".format(padding)

    def transpose_list_like(self, items, ref):
        return self.nxc_to_ncx(items) if ref is not None and self.transposing(ref) else items

    def transpose_axis_like(self, axis, ref, rank=None):
        return self.axis_nxc_to_ncx(axis, rank or ref.rank) if ref is not None and self.transposing(ref) else \
            self.ensure_positive(axis, rank or ref.rank)

    def squeeze_input(self, tensor, axes, keep_dims=False):
        return self._pre_squeeze(tensor, axes=axes) if not keep_dims else tensor

    def squeeze_output(self, tensor, axes, keep_dims=False):
        return self._post_squeeze(tensor, axes=axes) if not keep_dims else tensor

    def unsqueeze_input(self, tensor, axes, keep_dims=False):
        return self._pre_unsqueeze(tensor, axes=axes) if not keep_dims else tensor

    def unsqueeze_output(self, tensor, axes, keep_dims=False):
        return self._post_unsqueeze(tensor, axes=axes) if not keep_dims else tensor

    def unsqueeze_vector(self, tensor):
        original = self._tensor_map[tensor]
        if self._is_constant(original) and len(original.consumers) == 1:
            self._transform_constant(tensor, lambda data: np.expand_dims(data, 0))
            return tensor
        else:
            return self.unsqueeze_input(tensor, axes=[0])

    def convert_binarg(self, tensor, other):
        if tensor.rank == 0:
            return tensor
        needs_transpose = self.transposing(other) and not self.transposing(tensor)
        if other.rank > tensor.rank:
            if tensor.rank == 1 and needs_transpose:
                return self.unsqueeze_vector(tensor)
            tensor = self._pre_unsqueeze(tensor, axes=list(range(other.rank - tensor.rank)))
        return self.transpose_input(tensor) if needs_transpose else tensor

    def ensure_list(self, value):
        return value if isinstance(value, list) else list(value) if isinstance(value, tuple) else [value]

    def is_bit_set(self, mask, idx):
        return mask & (1 << idx) != 0

    def bit_count(self, mask):
        count = 0
        for i in range(mask.bit_length()):
            if self.is_bit_set(mask, i):
                count += 1
        return count

    def replace_item_with(self, items, index, count, value):
        return items[:index] + [value] * count + items[index+1:]

    def replace_bit_with(self, mask, index, count, value):
        value_bits = (((1 << count) - 1) << index) if value else 0
        low_bits = mask & ((1 << index) - 1)
        high_bits = (mask & ~((1 << (index + 1)) - 1)) << (count - 1)
        return low_bits | value_bits | high_bits

    def beg_index(self, stride):
        return _INT_MAX if stride < 0 else 0

    def end_index(self, stride):
        return _INT_MAX if stride > 0 else -_INT_MAX


_Transforms = Converter.unpack_transforms({
    'Placeholder':
        Transform(
            type='external',
            using={'needs_transpose': '!needs_io_transpose(O[0])'},
            outputs='!transpose_output(O[0]) if needs_transpose else O[0]',
            attribs={
                'shape': '!list(nxc_to_ncx(shape) if needs_transpose else shape)',
                'dtype': '!dtype',
            }
        ),
    'Const':
        Transform(
            type='constant',
            outputs='!O[0]',
            attribs={
                'shape': '!list(value.shape)',
                'dtype': '!dtype',
                'value': '!value',
            }
        ),
    ('Conv2D', 'Conv3D', 'DepthwiseConv2dNative'):
        Transform(
            type='conv',
            using={
                'depthwise': '!_type_ == "DepthwiseConv2dNative"',
            },
            defaults={
                'explicit_paddings': [],
            },
            inputs=(
                '!transpose_input(I[0], data_format)',
                '!transpose_filter(I[1]) if not depthwise else transpose_depthwise_filter(I[1])',
            ),
            outputs=(
                '!transpose_output(O[0], data_format)',
            ),
            attribs={
                'stride': '!convert_size(strides, data_format)',
                'dilation': '!convert_size(dilations, data_format)',
                'padding': '!convert_padding(padding, I[0].rank - 2, explicit_paddings, data_format)',
                'groups': '!1 if not depthwise else 0',
            }
        ),
    ('Conv2DBackpropInput', 'Conv3DBackpropInput', 'DepthwiseConv2dNativeBackpropInput'):
        Transform(
            type='deconv',
            using={
                'depthwise': '!_type_ == "DepthwiseConv2dNativeBackpropInput"',
            },
            defaults={
                'explicit_paddings': [],
            },
            inputs=(
                '!transpose_input(I[2], data_format)',
                '!transpose_filter(I[1]) if not depthwise else transpose_depthwise_filter(I[1])',
            ),
            outputs=(
                '!transpose_output(O[0], data_format)',
            ),
            attribs={
                'stride': '!convert_size(strides, data_format)',
                'dilation': '!convert_size(dilations, data_format)',
                'padding': '!convert_padding(padding, I[2].rank - 2, explicit_paddings, data_format)',
                'output_shape': '!nxc_to_ncx(as_const(I[0])) if is_nxc(data_format) else as_const(I[0])',
                'groups': '!1 if not depthwise else 0',
            }
        ),
    ('MaxPool', 'AvgPool', 'MaxPoolWithArgmax'):
        Transform(
            type=('max_pool', 'avg_pool', 'max_pool_with_index'),
            defaults={
                'explicit_paddings': [],
                'data_format': 'NHWC',
            },
            inputs=(
                '!transpose_input(I[0], data_format)',
            ),
            outputs=(
                '!transpose_output(O[0], data_format)',
                '!transpose_output(O[1], data_format) if len(O) > 1 else None',
            ),
            attribs={
                'size': '!nxc_to_ncx(ksize) if is_nxc(data_format) else ksize',
                'stride': '!nxc_to_ncx(strides) if is_nxc(data_format) else strides',
                'padding': '!convert_padding(padding, I[0].rank, explicit_paddings, data_format)',
            }
        ),
    'Concat':
        Transform(
            type='concat',
            inputs=['!I[1:]'],
            outputs='!transpose_like(O[0], I[1])',
            attribs={
                'axis': '!transpose_axis_like(as_const(I[0]), I[1], O[0].rank)',
            }
        ),
    'ConcatV2':
        Transform(
            type='concat',
            inputs=['!I[:-1]'],
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'axis': '!transpose_axis_like(as_const(I[-1]), I[0], O[0].rank)',
            }
        ),
    'Split':
        Transform(
            type='split',
            inputs='!I[1]',
            outputs=['![transpose_like(O[i], I[1]) for i in range(len(O))]'],
            attribs={
                'axis': '!transpose_axis_like(as_const(I[0]), I[1])',
                'ratios': '![1] * (num_split if not _lite_ else num_splits)',
            }
        ),
    'SplitV':
        Transform(
            type='split',
            inputs='!I[0]',
            outputs=['![transpose_like(O[i], I[0]) for i in range(len(O))]'],
            attribs={
                'axis': '!transpose_axis_like(as_const(I[2]), I[0])',
                'ratios': '!as_const(I[1])',
            }
        ),
    'Reshape':
        Transform(
            type='reshape',
            inputs='!undo_transpose(I[0])',
            outputs='!O[0]',
            attribs={
                'shape': '!flexible_batch(as_const(I[1]), I[0].shape[0])',
                'dtype': '!I[0].dtype',
            }
        ),
    'Transpose':
        Transform(
            type='transpose',
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axes': '!transpose_axis_like(as_const(I[1]), I[0])',
                'dtype': '!I[0].dtype',
            }
        ),
    'Squeeze':
        Transform(
            type='squeeze',
            inputs='!undo_transpose(I[0])',
            outputs='!O[0]',
            attribs={
                'axes': '!ensure_list(ensure_positive(squeeze_dims, I[0].rank)) if len(squeeze_dims) != 0 else'
                        ' [i for i, x in enumerate(I[0].shape) if x == 1]',
                'dtype': '!I[0].dtype',
            }
        ),
    'ExpandDims':
        Transform(
            type='unsqueeze',
            inputs='!undo_transpose(I[0])',
            outputs='!O[0]',
            attribs={
                'axes': '!ensure_list(ensure_positive(as_const(I[1]), O[0].rank))',
                'dtype': '!I[0].dtype',
            }
        ),
    'Pack':
        Transform(
            type='stack',
            inputs=['![undo_transpose(t) for t in I]'],
            outputs='!O[0]',
            attribs={
                'axis': '!ensure_positive(axis, O[0].rank)',
            }
        ),
    'Unpack':
        Transform(
            type='unstack',
            inputs='!undo_transpose(I[0])',
            outputs=['!O[:]'],
            attribs={
                'axis': '!ensure_positive(axis, I[0].rank)',
            }
        ),
    ('Min', 'Max', 'Mean', 'Sum', 'Any', 'All'):
        Transform(
            type=('min_reduce', 'max_reduce', 'mean_reduce', 'sum_reduce', 'any_reduce', 'all_reduce'),
            using={
                'axes': '!ensure_list(transpose_axis_like(as_const(I[1]), I[0]))'
            },
            inputs='!I[0]',
            outputs=(
                '!transpose_like(O[0], I[0]) if keep_dims else squeeze_output(O[0], axes)',
            ),
            attribs={
                'axes': '!axes',
            }
        ),
    ('Add', 'AddV2', 'Sub', 'Mul', 'RealDiv', 'Pow', 'LogicalAnd', 'LogicalOr',
     'Less', 'Greater', 'LessEqual', 'GreaterEqual', 'Equal', 'NotEqual', 'Minimum', 'Maximum'):
        Transform(
            type=('add', 'add', 'sub', 'mul', 'div', 'pow', 'and', 'or',
                  'lt', 'gt', 'le', 'ge', 'eq', 'ne', 'min', 'max'),
            inputs=(
                '!convert_binarg(I[0], I[1])',
                '!convert_binarg(I[1], I[0])',
            ),
            outputs='!transpose_output(O[0]) if transposing(I[0]) or transposing(I[1]) else O[0]',
        ),
    ('Identity', 'Relu', 'Relu6', 'Elu', 'Selu', 'Gelu', 'Silu', 'Swish', 'Sigmoid', 'Softplus', 'Exp', 'Log',
     'Sin', 'Cos', 'Tan', 'Asin', 'Acos', 'Atan', 'Sinh', 'Cosh', 'Tanh', 'Asinh', 'Acosh', 'Atanh',
     'Neg', 'Reciprocal', 'Sign', 'Abs', 'Floor', 'Ceil', 'Round', 'Square', 'Sqrt', 'Rsqrt', 'LogicalNot'):
        Transform(
            type=('copy', 'relu', 'relu6', 'elu', 'selu', 'gelu', 'silu', 'silu', 'sigmoid', 'softplus', 'exp', 'log',
                  'sin', 'cos', 'tan', 'asin', 'acos', 'atan', 'sinh', 'cosh', 'tanh', 'asinh', 'acosh', 'atanh',
                  'neg', 'rcp', 'sign', 'abs', 'floor', 'ceil', 'round', 'sqr', 'sqrt', 'rsqrt', 'not'),
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
        ),
    'LeakyRelu':
        Transform(
            type='leaky_relu',
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'alpha': '!alpha',
            }
        ),
    ('FusedBatchNorm', 'FusedBatchNormV3'):
        Transform(
            type='batch_normalization',
            inputs=(
                '!transpose_input(I[0], data_format)',
                '!unsqueeze_vector(I[3])',
                '!unsqueeze_vector(I[4])',
                '!unsqueeze_vector(I[2])',
                '!unsqueeze_vector(I[1])',
            ),
            outputs=(
                '!transpose_output(O[0], data_format)',
            ),
            attribs={
                'epsilon': '!epsilon',
            }
        ),
    'BiasAdd':
        Transform(
            type='add',
            inputs=(
                '!transpose_input(I[0], data_format)',
                '!unsqueeze_vector(I[1])',
            ),
            outputs='!transpose_output(O[0], data_format)',
        ),
    'Softmax':
        Transform(
            type='softmax',
            cond={
                '!beta == 1 if _lite_ else True': 'beta must be 1',
            },
            inputs='!transpose_input(I[0])',
            outputs='!transpose_output(O[0])',
            attribs={
                'axes': [1],
            }
        ),
    'MatMul':
        Transform(
            type='matmul',
            inputs=('!I[0]', '!I[1]'),
            outputs='!O[0]',
            attribs={
                'transposeA': '!transpose_a',
                'transposeB': '!transpose_b',
            },
        ),
    'ClipByValue':
        Transform(
            type='clamp',
            inputs=('!I[0]', '!I[1]', '!I[2]'),
            outputs='!transpose_like(O[0], I[0])',
        ),
    ('Pad', 'MirrorPad'):
        Transform(
            type='pad',
            cond={
                '!mode in ["CONSTANT", "REFLECT", "SYMMETRIC"]':
                    'mode must be one of "CONSTANT", "REFLECT" or SYMMETRIC',
            },
            defaults={
                'mode': 'CONSTANT',
            },
            using={
                'paddings': '!transpose_list_like(as_const(I[1]), ref=I[0])',
            },
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'padding': '![tuple(item) for item in paddings]',
                'border': '!"reflect" if mode == "REFLECT" else "reflect-even" if mode == "SYMMETRIC" else "constant"',
            }
        ),
    'Slice':
        Transform(
            type='slice',
            using={
                'beg': '!as_const(I[1])',
                'end': '![0 if s == -1 else b + s for b, s in zip(as_const(I[1]), as_const(I[2]))]',
            },
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'axes': '!list(range(I[0].rank))',
                'begin': '!transpose_list_like(beg, ref=I[0])',
                'end': '!transpose_list_like(end, ref=I[0])',
            }
        ),
    'StridedSlice':
        Transform(
            type='slice',
            using=OrderedDict([
                ('ref', '!I[0] if new_axis_mask == 0 else None'),
                ('rank', '!I[0].rank + bit_count(new_axis_mask)'),
                ('stride', '!as_const(I[3])'),
                ('ellipsis_index', '!int(math.log2(ellipsis_mask)) if ellipsis_mask != 0 else None'),
                ('ellipsis_count', '!rank - (len(stride) - 1) if ellipsis_mask != 0 else None'),
                ('beg', '!replace_item_with(as_const(I[1]), ellipsis_index, ellipsis_count, 0) '
                        'if ellipsis_index is not None else as_const(I[1])'),
                ('end', '!replace_item_with(as_const(I[2]), ellipsis_index, ellipsis_count, 0) '
                        'if ellipsis_index is not None else as_const(I[2])'),
                ('begin_mask', '!replace_bit_with(begin_mask, ellipsis_index, ellipsis_count, 1) '
                               'if ellipsis_index is not None else begin_mask'),
                ('end_mask', '!replace_bit_with(end_mask, ellipsis_index, ellipsis_count, 1) '
                             'if ellipsis_index is not None else end_mask'),
                ('new_axis_mask', '!replace_bit_with(new_axis_mask, ellipsis_index, ellipsis_count, 0) '
                                  'if ellipsis_index is not None else new_axis_mask'),
                ('shrink_axis_mask', '!replace_bit_with(shrink_axis_mask, ellipsis_index, ellipsis_count, 0) '
                                     'if ellipsis_index is not None else shrink_axis_mask'),
                ('masked_beg', '![beg_index(s) if is_bit_set(begin_mask,i) else b '
                               'for i, (b, s) in enumerate(zip(beg,stride))]'),
                ('masked_end', '![end_index(s) if is_bit_set(end_mask,i) else b + 1 '
                               'if is_bit_set(shrink_axis_mask,i) else e '
                               'for i, (b, e, s) in enumerate(zip(beg,end,stride))]'),
                ('axes', '!transpose_axis_like([i for i in range(rank) '
                         'if not (is_bit_set(begin_mask,i) and is_bit_set(end_mask,i)) '
                         'and not is_bit_set(new_axis_mask,i)], ref, rank)'),
                ('new_axes', '!transpose_axis_like([i for i in range(rank) '
                             'if is_bit_set(new_axis_mask,i)], ref, rank)'),
                ('del_axes', '!transpose_axis_like([i for i in range(rank) '
                             'if is_bit_set(shrink_axis_mask,i)], ref, rank)'),
            ]),
            inputs='!unsqueeze_input(undo_transpose(I[0]), new_axes) if len(new_axes) else I[0]',
            outputs='!transpose_like(squeeze_output(O[0], del_axes) if len(del_axes) else O[0], ref)',
            attribs={
                'axes': '![i for i in range(rank) if i in axes]',
                'begin': '![b for i, b in enumerate(transpose_list_like(masked_beg, ref)) if i in axes]',
                'end': '![e for i, e in enumerate(transpose_list_like(masked_end, ref)) if i in axes]',
                'stride': '![s for i, s in enumerate(transpose_list_like(stride, ref)) if i in axes]',
            }
        ),
    ('ArgMin', 'ArgMax'):
        Transform(
            type=('argmin_reduce', 'argmax_reduce'),
            using={
                'axis': '!transpose_axis_like(as_const(I[1]), ref=I[0])'
            },
            inputs='!I[0]',
            outputs='!transpose_like(O[0], ref=I[0]) if _lite_ else squeeze_output(O[0], [axis])',
            attribs={
                'axes': '!ensure_list(axis)',
            }
        ),
    'Select':
        Transform(
            type='select',
            inputs=(
                '!I[0]',
                '!convert_binarg(I[1], I[0])',
                '!convert_binarg(I[2], I[0])',
            ),
            outputs='!transpose_like(O[0], ref=I[0])',
        ),
    'Tile':
        Transform(
            type='tile',
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'repeats': '!transpose_list_like(as_const(I[1]), I[0])',
            }
        ),
    'ResizeNearestNeighbor':
        Transform(
            type='!"nearest_upsample" if upsample else "nearest_downsample"',
            using=OrderedDict([
                ('old_size', '!I[0].shape[1:-1]'),
                ('new_size', '!as_const(I[1])'),
                ('upsample', '!is_integer_upsample(old_size, new_size)'),
                ('downsample', '!is_integer_downsample(old_size, new_size)'),
            ]),
            cond={
                '!upsample or downsample': 'nearest resize must be integer up-sample or down-sample',
                '!not align_corners': 'align_corners is not supported',
            },
            inputs='!transpose_input(I[0])',
            outputs='!transpose_output(O[0])',
            attribs={
                'factor': '!upsample_factor(old_size, new_size) if upsample else downsample_factor(old_size, new_size)',
            }
        ),
    'ResizeArea':
        Transform(
            type='area_downsample',
            cond={
                '!is_integer_downsample(I[0].shape[1:-1], O[0].shape[1:-1])': 'area resize must be integer down-sample',
                '!not align_corners': 'align_corners is not supported',
            },
            using={
                'size': '!I[0].shape[1:-1]'
            },
            inputs='!transpose_input(I[0])',
            outputs='!transpose_output(O[0])',
            attribs={
                'factor': '!downsample_factor(size, as_const(I[1]))',
            }
        ),
    'ResizeBilinear':
        Transform(
            type='multilinear_upsample',
            cond={
                '!is_integer_upsample(I[0].shape[1:-1], O[0].shape[1:-1])': 'bilinear resize must be integer up-sample',
            },
            using={
                'size': '!I[0].shape[1:-1]'
            },
            inputs='!transpose_input(I[0])',
            outputs='!transpose_output(O[0])',
            attribs={
                'factor': '!upsample_factor(size, as_const(I[1]))',
                'method': '!"aligned" if align_corners else "symmetric" if half_pixel_centers else "asymmetric"',
            }
        ),
    'LRN':
        Transform(
            type='local_response_normalization',
            using={
                'size': '!(radius if _lite_ else depth_radius) * 2 + 1'
            },
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'size': '![1, size] + [1] * (I[0].rank - 2)',
                'alpha': '!alpha * size',
                'beta': '!beta',
                'bias': '!bias',
            }
        ),
    'Cast':
        Transform(
            using={
                'same_type': '!nnef_dtype(O[0].dtype) == nnef_dtype(I[0].dtype)',
            },
            type='!"copy" if same_type else "cast"',
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'dtype': '!O[0].dtype if not same_type else None',
            },
        ),
    ('Gather', 'GatherV2'):
        Transform(
            type='gather',
            using={
                'axis': '!transpose_axis_like(as_const(I[2]), ref=I[0])'
            },
            inputs=('!I[0]', '!I[1]'),
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'axis': '!axis',
            },
        ),
    'AddN':
        Transform(
            type='add_n',
            inputs=['!I[:]'],
            outputs='!transpose_like(O[0], I[0])'
        ),
})



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import
from .converter import Converter as _Converter, Transform, ConversionError
from .tf_to_nnef import Converter as _TFConverter, _Transforms as _TFTransforms, _RELU6_FRAGMENT
from ..model import Tensor, Operation
from ..utils import types
from ..io.tf.lite import CustomOptionsKey
import numpy as np
import copy


_DETECTION_POSTPROCESS_FRAGMENT = """
fragment TFLite_Detection_PostProcess( 
    boxes: tensor<scalar>, 
    scores: tensor<scalar>, 
    anchors: tensor<scalar>, 
    detections_per_class: integer,
    max_classes_per_detection: integer,
    max_detections: integer, 
    nms_iou_threshold: scalar, 
    nms_score_threshold: scalar, 
    num_classes: integer, 
    use_regular_nms: logical, 
    h_scale: scalar,
    w_scale: scalar, 
    x_scale: scalar, 
    y_scale: scalar ) 
-> ( 
    detection_boxes: tensor<scalar>, 
    detection_classes: tensor<scalar>, 
    detection_scores: tensor<scalar>, 
    num_detections: tensor<scalar> );
"""


class Converter(_TFConverter):

    _ConvOpTypes = ['CONV_1D', 'CONV_2D', 'CONV_3D', 'TRANSPOSE_CONV', 'DEPTHWISE_CONV_2D']

    _ActivationOpTypes = {
        'ELU': 'elu',
        'RELU': 'relu',
        'RELU6': 'relu6',
        'LOGISTIC': 'sigmoid',
        'TANH': 'tanh',
    }

    @staticmethod
    def defined_operations():
        return {
            'relu6': _RELU6_FRAGMENT,
            'TFLite_Detection_PostProcess': _DETECTION_POSTPROCESS_FRAGMENT,
        }

    def __init__(self, io_transpose=False, custom_transforms=None, custom_functions=None,
                 mirror_unsupported=False, keep_io_names=False):
        _Converter.__init__(self, transforms=self.merge_transforms(_Transforms, custom_transforms),
                            functions=custom_functions, mirror_unsupported=mirror_unsupported)
        self._io_transpose = io_transpose
        self._keep_io_names = keep_io_names

    def __call__(self, graph):
        graph = _TFConverter.__call__(self, graph)
        self._fix_custom_options(graph)
        return graph

    def _global_attribs(self):
        return {'_lite_': True}

    def _prepare(self, graph):
        self._fix_quantization_attribs(graph)
        self._fix_quantized_dtypes(graph)
        self._insert_externals_and_constants(graph)
        self._transpose_externals(graph)

    def _is_constant(self, tensor):
        return tensor.producer is None and tensor.data is not None

    def _read_constant(self, tensor, type=None):
        if tensor.producer is None:
            return types.from_numpy(tensor.data, type=type)
        else:
            raise ConversionError('trying to evaluate non-constant tensor')

    def _transpose_externals(self, graph):
        for op in graph.operations:
            if op.type == 'external':
                if self.needs_io_transpose(op.output):
                    shape = self.nxc_to_ncx(op.output.shape)
                    op.attribs['shape'] = list(shape)
                    self._transposes[op.output] = shape

    @staticmethod
    def _is_zero(value):
        return np.all(value == 0) if isinstance(value, np.ndarray) else value == 0

    def _fix_quantized_dtypes(self, graph):
        for tensor in graph.tensors:
            if tensor.quant:
                scale = tensor.quant.get('scale')
                if scale is not None and not self._is_zero(scale):
                    tensor.dtype = np.float32
                else:
                    tensor.quant = None

    def _fix_quantization_attribs(self, graph):
        dtype_bits = {
            np.int8: 8,
            np.uint8: 8,
            np.int16: 16,
            np.uint16: 16,
            np.int32: 32,
            np.uint32: 32,
            np.int64: 64,
            np.uint64: 64,
        }

        for tensor in graph.tensors:
            if tensor.quant:
                scale = tensor.quant.get('scale')
                zero_point = tensor.quant.get('zero_point')
                if scale is not None and not self._is_zero(scale):
                    if 'min' in tensor.quant:
                        del tensor.quant['min']
                    if 'max' in tensor.quant:
                        del tensor.quant['max']
                    assert tensor.dtype == np.uint8 or tensor.dtype == np.int8 or \
                           tensor.dtype == np.uint16 or tensor.dtype == np.int16 or \
                           tensor.dtype == np.uint32 or tensor.dtype == np.int32, \
                        "unknown quantized dtype '{}'".format(tensor.dtype)
                    tensor.quant['op-name'] = 'zero_point_linear_quantize'
                    tensor.quant['bits'] = 32 if self._is_conv_bias(tensor) else dtype_bits[tensor.dtype]
                    tensor.quant['signed'] = tensor.dtype == np.int8 or tensor.dtype == np.int16 or tensor.dtype == np.int32
                    tensor.quant['symmetric'] = self._is_conv_filter(tensor)

                    if tensor.data is None:
                        if isinstance(zero_point, np.ndarray) and len(zero_point.shape) == 1:
                            tensor.quant['zero_point'] = np.expand_dims(zero_point, axis=0)
                        if isinstance(scale, np.ndarray) and len(scale.shape) == 1:
                            tensor.quant['scale'] = np.expand_dims(scale, axis=0)

    def _fix_custom_options(self, graph):
        for op in graph.operations:
            if op.custom:
                options = op.attribs.get(CustomOptionsKey)
                if options is not None:
                    op.attribs[CustomOptionsKey] = options.hex()

    def _is_conv_filter(self, tensor):
        tensor = self._tensor_map.get(tensor)
        return tensor and len(tensor.consumers) > 0 and \
               all(op.type in Converter._ConvOpTypes and op.inputs[1] is tensor for op in tensor.consumers)

    def _is_conv_bias(self, tensor):
        tensor = self._tensor_map.get(tensor)
        return tensor and len(tensor.consumers) > 0 and \
               all(op.type in Converter._ConvOpTypes and op.inputs[2] is tensor for op in tensor.consumers)

    def activation(self, output, func):
        if func is None or func == 'NONE':
            return output

        if func not in self._ActivationOpTypes:
            raise ConversionError("Unsupported fused activation function '{}'".format(func))

        input = Tensor(output.graph, dtype=output.dtype, shape=self._working_shape(output), quant=copy.deepcopy(output.quant))
        Operation(output.graph, type=self._ActivationOpTypes[func], inputs=input, outputs=output)
        return input

    def flat_list(self, array):
        return [item for items in array for item in items] if len(array) and isinstance(array[0], (list, tuple)) else array

    def flatten(self, input):
        shape = (input.shape[0], int(np.prod(input.shape[1:])))
        output = Tensor(input.graph, dtype=input.dtype, shape=shape, quant=copy.deepcopy(input.quant))
        self._reshape_operation(input, output, shape)
        return output

    def same_shape(self, input, output):
        return self._tensor_map[input].shape == self._tensor_map[output].shape


_Transforms = Converter.unpack_transforms({
    ('CONV_1D', 'CONV_2D', 'CONV_3D', 'DEPTHWISE_CONV_2D'):
        Transform(
            type='conv',
            using={
                'depthwise': '!_type_ == "DEPTHWISE_CONV_2D"',
            },
            inputs=(
                '!transpose_input(I[0])',
                '!transpose_filter(I[1], format="NXC" if not depthwise else "CXN")',
                '!unsqueeze_vector(I[2])',
            ),
            outputs='!activation(transpose_output(O[0]), fused_activation_function)',
            attribs={
                'stride': '![stride_h, stride_w]',
                'dilation': '![dilation_h_factor, dilation_w_factor]',
                'padding': '!convert_padding(padding, I[0].rank - 2)',
                'groups': '!1 if not depthwise else 0',
            }
        ),
    'TRANSPOSE_CONV':
        Transform(
            type='deconv',
            using={
                'depthwise': False,
            },
            inputs=(
                '!transpose_input(I[2])',
                '!transpose_filter(I[1], format="CXN" if not depthwise else "NXC")',
            ),
            outputs='!transpose_output(O[0])',
            attribs={
                'stride': '![stride_h, stride_w]',
                'padding': '!convert_padding(padding, I[0].rank - 2)',
                'output_shape': '!nxc_to_ncx(as_const(I[0]))',
                'groups': '!1 if not depthwise else 0',
            }
        ),
    ('MAX_POOL_2D', 'AVERAGE_POOL_2D'):
        Transform(
            type=('max_pool', 'avg_pool'),
            inputs=(
                '!transpose_input(I[0])',
            ),
            outputs=(
                '!transpose_output(O[0])',
            ),
            attribs={
                'size': '![1, 1, filter_height, filter_width]',
                'stride': '![1, 1, stride_h, stride_w]',
                'padding': '!convert_padding(padding, I[0].rank)',
            }
        ),
    'RESHAPE':
        Transform(
            type='reshape',
            inputs='!undo_transpose(I[0])',
            outputs='!O[0]',
            attribs={
                'shape': '!flexible_batch(flat_list(as_const(I[1])) if len(I) > 1 else new_shape, I[0].shape[0])',
                'dtype': '!I[0].dtype',
            }
        ),
    'CONCATENATION':
        Transform(
            type='concat',
            inputs=['!I[:]'],
            outputs='!activation(transpose_like(O[0], I[0]), fused_activation_function)',
            attribs={
                'axis': '!transpose_axis_like(axis, I[0], O[0].rank)',
            }
        ),
    'FULLY_CONNECTED':
        Transform(
            type='linear',
            cond={
                '!weights_format == "DEFAULT"': 'wights_format must be "DEFAULT"',
            },
            inputs=(
                '!I[0] if keep_num_dims else flatten(I[0])',
                '!I[1]',
                '!unsqueeze_vector(I[2]) if len(I) > 2 else None',
            ),
            outputs='!activation(O[0], fused_activation_function)',
        ),
    'BATCH_MATMUL':
        Transform(
            type='matmul',
            cond={
                '!asymmetric_quantize_inputs == False': 'asymmetric_quantize_inputs must be False',
            },
            inputs=('!I[0]', '!I[1]'),
            outputs='!O[0]',
            attribs={
                'transposeA': '!adj_x',
                'transposeB': '!adj_y',
            },
        ),
    'L2_NORMALIZATION':
        Transform(
            type='l2_normalization',
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'axes': '!list(range(I[0].rank))',
            }
        ),
    'PRELU':
        Transform(
            type='prelu',
            inputs=('!I[0]', '!I[1]'),
            outputs='!transpose_like(O[0], I[0])',
        ),
    ('PAD', 'MIRROR_PAD'):
        Transform(
            type='pad',
            cond={
                '!mode < 2': 'mode must be 0 or 1',
            },
            defaults={
                'mode': None,
            },
            using={
                'paddings': '!transpose_list_like(as_const(I[1]), ref=I[0])',
            },
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'padding': '![tuple(item) for item in paddings]',
                'border': '!"reflect" if mode == 0 else "reflect-even" if mode == 1 else "constant"',
            }
        ),
    'GATHER':
        Transform(
            type='gather',
            inputs=('!I[0]', '!I[1]'),
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'axis': '!transpose_axis_like(axis, I[0])',
            },
        ),
    'IDENTITY': _TFTransforms['Identity'],
    'QUANTIZE': _TFTransforms['Identity'],
    'TRANSPOSE': _TFTransforms['Transpose'],
    'SPLIT': _TFTransforms['Split'],
    'SPLIT_V': _TFTransforms['SplitV'],
    'PACK': _TFTransforms['Pack'],
    'UNPACK': _TFTransforms['Unpack'],
    'TILE': _TFTransforms['Tile'],
    'SQUEEZE': _TFTransforms['Squeeze'],
    'EXPAND_DIMS': _TFTransforms['ExpandDims'],
    'SLICE': _TFTransforms['Slice'],
    'STRIDED_SLICE': _TFTransforms['StridedSlice'],
    'RELU': _TFTransforms['Relu'],
    'RELU6': _TFTransforms['Relu6'],
    'ELU': _TFTransforms['Elu'],
    'LEAKY_RELU': _TFTransforms['LeakyRelu'],
    'LOGISTIC': _TFTransforms['Sigmoid'],
    'SIN': _TFTransforms['Sin'],
    'COS': _TFTransforms['Cos'],
    'TAN': _TFTransforms['Tan'],
    'ASIN': _TFTransforms['Asin'],
    'ACOS': _TFTransforms['Acos'],
    'ATAN': _TFTransforms['Atan'],
    'SINH': _TFTransforms['Sinh'],
    'COSH': _TFTransforms['Cosh'],
    'TANH': _TFTransforms['Tanh'],
    'ASINH': _TFTransforms['Asinh'],
    'ACOSH': _TFTransforms['Acosh'],
    'ATANH': _TFTransforms['Atanh'],
    'EXP': _TFTransforms['Exp'],
    'LOG': _TFTransforms['Log'],
    'ABS': _TFTransforms['Abs'],
    'NEG': _TFTransforms['Neg'],
    'LOGICAL_NOT': _TFTransforms['LogicalNot'],
    'FLOOR': _TFTransforms['Floor'],
    'CEIL': _TFTransforms['Ceil'],
    'ROUND': _TFTransforms['Round'],
    'SQUARE': _TFTransforms['Square'],
    'SQRT': _TFTransforms['Sqrt'],
    'RSQRT': _TFTransforms['Rsqrt'],
    'ADD': _TFTransforms['Add'],
    'SUB': _TFTransforms['Sub'],
    'MUL': _TFTransforms['Mul'],
    'DIV': _TFTransforms['RealDiv'],
    'POW': _TFTransforms['Pow'],
    'MINIMUM': _TFTransforms['Minimum'],
    'MAXIMUM': _TFTransforms['Maximum'],
    'LOGICAL_AND': _TFTransforms['LogicalAnd'],
    'LOGICAL_OR': _TFTransforms['LogicalOr'],
    'LESS': _TFTransforms['Less'],
    'LESS_EQUAL': _TFTransforms['LessEqual'],
    'GREATER': _TFTransforms['Greater'],
    'GREATER_EQUAL': _TFTransforms['GreaterEqual'],
    'EQUAL': _TFTransforms['Equal'],
    'NOT_EQUAL': _TFTransforms['NotEqual'],
    'SELECT': _TFTransforms['Select'],
    'REDUCE_MIN': _TFTransforms['Min'],
    'REDUCE_MAX': _TFTransforms['Max'],
    'MEAN': _TFTransforms['Mean'],
    'SUM': _TFTransforms['Sum'],
    'REDUCE_ANY': _TFTransforms['Any'],
    'REDUCE_ALL': _TFTransforms['All'],
    'ARG_MIN': _TFTransforms['ArgMin'],
    'ARG_MAX': _TFTransforms['ArgMax'],
    'SOFTMAX': _TFTransforms['Softmax'],
    'LOCAL_RESPONSE_NORMALIZATION': _TFTransforms['LRN'],
    'RESIZE_NEAREST_NEIGHBOR': _TFTransforms['ResizeNearestNeighbor'],
    'RESIZE_BILINEAR': _TFTransforms['ResizeBilinear'],
    'ADD_N': _TFTransforms['AddN'],
    'CAST': _TFTransforms['Cast'],
})



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...model import *
import numpy as np
import six
import onnx


_DtypeFromNumpy = {
    None: 'UNDEFINED',
    np.float32: 'FLOAT',
    np.uint8: 'UINT8',
    np.int8: 'INT8',
    np.uint16: 'UINT16',
    np.int16: 'INT16',
    np.int32: 'INT32',
    np.int64: 'INT64',
    np.str_: 'STRING',
    np.bool_: 'BOOL',
    np.float16: 'FLOAT16',
    np.float64: 'DOUBLE',
    np.uint32: 'UINT32',
    np.uint64: 'UINT64',
    np.complex64: 'COMPLEX64',
    np.complex128: 'COMPLEX128',
}


def build_model(graph, ir_version, opset_version):
    # type: (Graph)->onnx.ModelProto

    model_proto = onnx.ModelProto()
    build_graph(graph, model_proto.graph)

    model_proto.ir_version = ir_version
    model_proto.opset_import.add()
    model_proto.opset_import[0].version = opset_version

    return model_proto


def build_graph(graph, graph_proto):
    # type: (Graph, onnx.GraphProto)->None

    for idx, op in enumerate(graph.operations):
        node_proto = graph_proto.node.add()
        build_node(op, node_proto, idx)

    if graph.name is not None:
        graph_proto.name = graph.name

    for input in list(graph.inputs) + list(t for t in graph.tensors if t.is_constant and t.name != ''):
        value_info_proto = graph_proto.input.add()
        build_value_info(input, value_info_proto)

    for output in graph.outputs:
        value_info_proto = graph_proto.output.add()
        build_value_info(output, value_info_proto)

    for tensor in graph.tensors:
        if tensor.is_constant and tensor.name != '':
            tensor_proto = graph_proto.initializer.add()
            build_tensor_proto(tensor, tensor_proto)

        if tensor.quant:
            build_quantization(tensor, graph_proto)


def build_value_info(tensor, value_info_proto):
    # type: (Tensor, onnx.ValueInfoProto)->None

    value_info_proto.name = tensor.name
    value_info_proto.type.tensor_type.elem_type = build_dtype(tensor.dtype)

    if tensor.shape:
        for s in tensor.shape:
            dim = value_info_proto.type.tensor_type.shape.dim.add()
            if s is not None:
                dim.dim_value = s
    else:
        value_info_proto.type.tensor_type.shape.SetInParent()


def build_dtype(dtype):
    dtype = dtype.type if isinstance(dtype, np.dtype) else dtype
    return onnx.TensorProto.DataType.Value(_DtypeFromNumpy[dtype])


def build_attribute_type(name):
    return onnx.AttributeProto.AttributeType.Value(name)


def build_tensor_data(data, tensor_proto):
    # type: (np.ndarray, onnx.TensorProto)->None
    for s in data.shape:
        tensor_proto.dims.append(s)

    tensor_proto.data_type = build_dtype(data.dtype)

    if data.dtype == np.str_:
        tensor_proto.string_data = str(data)
    else:
        data = data.flatten().astype(data.dtype)

        if data.dtype in [np.complex64, np.complex128]:
            data = np.column_stack((np.real(data), np.imag(data))).flatten()

        if data.dtype.str[0] != "<":
            data = data.byteswap()
        tensor_proto.raw_data = data.tobytes()


def build_tensor_proto(tensor, tensor_proto):
    # type: (Tensor, onnx.TensorProto)->None
    if isinstance(tensor.data, np.ndarray):
        data = tensor.data
    elif isinstance(tensor.data, (list, tuple)):
        data = np.array(tensor.data, dtype=tensor.dtype).reshape(tensor.shape)
    else:
        data = np.full(shape=tensor.shape, fill_value=tensor.data, dtype=tensor.dtype)

    build_tensor_data(data, tensor_proto)
    tensor_proto.name = tensor.name


def build_quantization(tensor, graph_proto):
    tensor_annotation = graph_proto.quantization_annotation.add()
    tensor_annotation.tensor_name = tensor.name

    for key, value in tensor.quant:
        value_tensor_name = tensor.name + '/' + key
        if not isinstance(value, np.ndarray):
            value = np.array(value)

        tensor_proto = graph_proto.initializer.add()
        build_tensor_data(value, tensor_proto)
        tensor_proto.name = value_tensor_name

        item = tensor_annotation.quant_parameter_tensor_names.add()
        item.key = key
        item.value = value_tensor_name


def build_node(op, node_proto, idx):
    # type: (Operation, onnx.NodeProto)->None

    inputs = op.inputs
    attribs = op.attribs

    for input in inputs:
        node_proto.input.append(input.name)
    for output in op.outputs:
        node_proto.output.append(output.name)

    node_proto.op_type = op.type
    node_proto.name = op.name or (op.type + str(idx))

    for k, v in six.iteritems(attribs):
        attribute_proto = node_proto.attribute.add()
        build_attribute(k, v, attribute_proto)


def build_attribute(key, value, attribute_proto):
    # type: (str, typing.Any, onnx.AttributeProto)->None

    attribute_proto.name = key

    if isinstance(value, np.ndarray):
        attribute_proto.type = build_attribute_type('TENSOR')
        build_tensor_data(value, attribute_proto.t)
    elif isinstance(value, int):
        attribute_proto.type = build_attribute_type('INT')
        attribute_proto.i = value
    elif isinstance(value, float):
        attribute_proto.type = build_attribute_type('FLOAT')
        attribute_proto.f = value
    elif isinstance(value, str):
        attribute_proto.type = build_attribute_type('STRING')
        attribute_proto.s = value.encode('utf-8')
    elif isinstance(value, (type, np.dtype)):
        attribute_proto.type = build_attribute_type('INT')
        attribute_proto.i = build_dtype(value)
    elif isinstance(value, Graph):
        attribute_proto.type = build_attribute_type('GRAPH')
        build_graph(value, attribute_proto.g)
    elif isinstance(value, (list, tuple)):
        if len(value) == 0:
            attribute_proto.type = build_attribute_type('INTS')  # TODO better
        else:
            if isinstance(value[0], int):
                attribute_proto.type = build_attribute_type('INTS')
                for v in value:
                    attribute_proto.ints.append(v)
            elif isinstance(value[0], float):
                attribute_proto.type = build_attribute_type('FLOATS')
                for v in value:
                    attribute_proto.floats.append(v)
            elif isinstance(value[0], str):
                attribute_proto.type = build_attribute_type('STRINGS')
                for v in value:
                    attribute_proto.strings.append(v.encode('utf-8'))
            elif isinstance(value[0], Graph):
                attribute_proto.type = build_attribute_type('GRAPHS')
                for v in value:
                    g = attribute_proto.graphs.add()
                    build_graph(v, g)
            else:
                assert False, \
                    "Unsupported attribute: {}: {} of type: List[{}]".format(key, value, type(value[0]).__name__)
    else:
        assert False, "Unsupported attribute: {}: {} of type: {}".format(key, value, type(value).__name__)


class Writer(object):

    def __init__(self, ir_version=6, opset_version=11):
        self._ir_version = ir_version
        self._opset_version = opset_version

    def __call__(self, graph, filename):
        model_proto = build_model(graph, self._ir_version, self._opset_version)
        onnx.checker.check_model(model_proto)
        with open(filename, 'wb') as file:
            file.write(model_proto.SerializeToString())



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .utils import stdio
import numpy as np
import argparse
import nnef
import sys


def _is_lambda(v):
    LAMBDA = lambda: 0
    return isinstance(v, type(LAMBDA)) and v.__name__ == LAMBDA.__name__


def uniform(min=0, max=1):
    return lambda shape: np.random.uniform(min, max, shape)


def normal(mean=0, std=1):
    return lambda shape: np.random.normal(mean, std, shape)


def bernoulli(prob=0.5):
    return lambda shape: np.random.uniform(0, 1, shape) < prob


def main(args):
    if args.output is None:
        if not stdio.is_stdout_piped():
            print("Output must be piped", file=sys.stderr)
            return -1
        stdio.set_stdout_to_binary()

    try:
        distribution = eval(args.distribution)
        if not _is_lambda(distribution):
            distribution = distribution()
    except Exception as e:
        print("Could not evaluate distribution: " + str(e), file=sys.stderr)
        return -1

    tensor = distribution(args.shape).astype(np.dtype(args.dtype))

    if args.output is not None:
        with open(args.output, 'wb') as file:
            nnef.write_tensor(file, tensor)
    else:
        nnef.write_tensor(sys.stdout, tensor)
        
    return 0


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('distribution', type=str,
                        help='The distribution to generate values from')
    parser.add_argument('--shape', type=int, nargs='+', required=True,
                        help='The dimensions of the tensor to generate')
    parser.add_argument('--dtype', type=str, default='float32',
                        help='The data-type of the resulting tensor')
    parser.add_argument('--output', type=str, default=None,
                        help='File name to save the result into')
    exit(main(parser.parse_args()))



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import

import nnef
import numpy as np
import tempfile
import shutil
import six
import os
from .helpers import tgz_compress
from ...model import Tensor
from ...utils.types import as_str, from_numpy


_DtypeFromNumpy = {
    np.float16: 'scalar',
    np.float32: 'scalar',
    np.float64: 'scalar',
    np.int8: 'integer',
    np.uint8: 'integer',
    np.int16: 'integer',
    np.uint16: 'integer',
    np.int32: 'integer',
    np.uint32: 'integer',
    np.int64: 'integer',
    np.uint64: 'integer',
    np.bool_: 'logical',
}


_DtypeFromPyType = {
    str: 'string',
    float: 'scalar',
    int: 'integer',
    bool: 'logical',
    None: 'dtype',
}


def _nnef_dtype(dtype):
    return _DtypeFromNumpy[dtype.type if isinstance(dtype, np.dtype) else dtype] if dtype is not None else None


def _print(graph, file, extensions, fragments, version_custom_ops, annotate_shapes):
    assert graph.is_sorted(), "graph must be topologically sorted"
    assert all(tensor.name is not None or (tensor.producer is None and tensor.data is not None)
               for tensor in graph.tensors), \
        "all tensors must have names"
    assert all(all(s is not None for s in op.attribs['shape'])
               for op in graph.operations if op.type == 'external'), \
        "external ops must not contain undefined shapes"

    print(nnef.format_version((1, 0)), file=file)
    if len(extensions):
        print(file=file)
        print(nnef.format_extensions(extensions), file=file)
    if fragments:
        print(file=file)
        print(fragments, file=file)
    print(file=file)

    graph_name = as_str(graph.name) if graph.name is not None else "G"
    graph_inputs = [as_str(item.name) for item in graph.inputs]
    graph_outputs = [as_str(item.name) for item in graph.outputs]

    print("graph {}({}) -> ({})".format(graph_name, ', '.join(graph_inputs), ', '.join(graph_outputs)), file=file)
    print("{", file=file)

    versions = {}
    for op in graph.operations:
        assert all(isinstance(item, Tensor) for item in op.outputs)

        inputs = ((from_numpy(item.data) if item.producer is None else nnef.Identifier(as_str(item.name)))
                  if isinstance(item, Tensor) else item for item in op.inputs)
        inputs = tuple(inputs) if isinstance(op.inputs, tuple) else (list(inputs),)

        outputs = (nnef.Identifier(as_str(item.name)) for item in op.outputs)
        outputs = tuple(outputs) if isinstance(op.outputs, tuple) else (list(outputs),)

        attribs = {as_str(key): value for key, value in six.iteritems(op.attribs)}

        name = _next_version(op.type, versions) if op.type not in nnef.StandardOperations and version_custom_ops else op.type

        dtype = attribs.get('dtype')
        if dtype is not None:
            dtype = _nnef_dtype(dtype)
            del attribs['dtype']

        for key, value in six.iteritems(attribs):
            if isinstance(value, (type, np.dtype)):
                attribs[key] = _nnef_dtype(value)

        invocation = nnef.format_invocation(name=name, dtype=dtype, attribs=attribs, inputs=inputs, outputs=outputs)
        annotation = "    # " + ", ".join(_nnef_dtype(output.dtype) + str(output.shape) for output in op.outputs) \
            if annotate_shapes else ''

        print("    {};{}".format(invocation, annotation), file=file)

    print("}", file=file)


def _write_tensor(array, filename, quantized):
    directory = os.path.dirname(filename)
    if directory and not os.path.exists(directory):
        os.makedirs(directory)

    with open(filename, "wb") as file:
        nnef.write_tensor(file=file, tensor=array, quantized=quantized)


def _write_quantization(graph, file):
    for tensor in graph.tensors:
        if tensor.quant:
            op_name = tensor.quant['op-name']
            attribs = ', '.join("{} = {}".format(k, _printable_value(v))
                                for k, v in six.iteritems(tensor.quant)
                                if k != 'op-name' and v is not None)
            if attribs:
                print('"{}": {}({});'.format(tensor.name, op_name, attribs), file=file)


def _printable_value(v):
    if type(v) == bool:
        return 'true' if v else 'false'
    elif type(v) == np.ndarray:
        return v.tolist()
    else:
        return v


def _next_version(name, versions):
    version = versions.get(name, 0) + 1
    versions[name] = version
    return '{}_v{}'.format(name, version)


def _generate_custom_fragments(graph, fragments, version):
    versions = {} if version else None
    return '\n'.join(_generate_fragment(op, versions) for op in graph.operations
                     if op.type not in nnef.StandardOperations and op.type not in fragments)


def _generate_fragment(op, versions):
    attribs = {key: _make_attrib_type(value) for key, value in op.attribs.items()}
    inputs = [_make_tensor_type(value) for value in op.inputs]
    outputs = [_make_tensor_type(value) for value in op.outputs]
    dtype = _nnef_dtype(op.attribs.get('dtype'))
    name = _next_version(op.type, versions) if versions is not None else op.type

    return 'fragment ' + _fragment_signature(name, dtype, attribs, inputs, outputs) + ';'


def _fragment_signature(name, dtype, attribs, inputs, outputs):
    str = name
    if dtype is not None:
        str += '<' + dtype + '>'
    str += '( '
    str += _types_str(['_I{}'.format(i + 1) for i in range(len(inputs))], inputs, True)
    if len(inputs) and len(attribs):
        str += ', '
    str += _types_str(attribs.keys(), attribs.values(), False)
    str += ' ) -> ( '
    str += _types_str(['_O{}'.format(i + 1) for i in range(len(outputs))], outputs, True)
    str += ' )'
    return str


def _make_attrib_type(value):
    repeated = False
    if isinstance(value, list):
        if len(value) == 0:
            return None, False
        tp = type(value[0])
        if not all(type(v) == tp for v in value):
            return None, False
        repeated = True
        value = value[0]

    if not isinstance(value, (float, int, bool, str)):
        return None, False

    return _DtypeFromPyType[type(value)], repeated


def _make_tensor_type(value):
    repeated = False
    if isinstance(value, list):
        if len(value) == 0:
            return None, False
        dtype = value[0].dtype
        if not all(v.dtype == dtype for v in value):
            return None, False
        repeated = True
        value = value[0]

    return _nnef_dtype(value.dtype), repeated


def _types_str(names, items, tensor):
    return ', '.join(name + ': ' + ('tensor<{}>'.format(type) if tensor else type) + ('[]' if repeated else '')
                     for name, (type, repeated) in zip(names, items))


class Writer(object):

    def __init__(self, compression=None, extensions=None, fragments=None, fragment_dependencies=None,
                 generate_custom_fragments=False, version_custom_fragments=True, annotate_shapes=False):
        self._compression = compression
        self._extensions = extensions or []
        self._fragments = fragments or {}
        self._fragment_dependencies = fragment_dependencies or {}
        self._generate_custom_fragments = generate_custom_fragments
        self._version_custom_fragments = version_custom_fragments
        self._annotate_shapes = annotate_shapes

    def __call__(self, graph, path):
        folder = None
        try:
            if self._compression is not None:
                folder = tempfile.mkdtemp(prefix="nnef_")
            else:
                folder = path
                if not os.path.exists(folder):
                    os.makedirs(folder)

            used_operators = self._used_operators(graph, self._fragment_dependencies)
            fragments = "".join(text for name, text in six.iteritems(self._fragments) if name in used_operators)
            if self._generate_custom_fragments:
                customs = _generate_custom_fragments(graph, fragments=self._fragments,
                                                     version=self._version_custom_fragments)
                if fragments and customs:
                    fragments += "\n"
                fragments += customs

            if len(fragments):
                if "KHR_enable_fragment_definitions" not in self._extensions:
                    self._extensions.append("KHR_enable_fragment_definitions")
                if "KHR_enable_operator_expressions" not in self._extensions:
                    self._extensions.append("KHR_enable_operator_expressions")

            graph_filename = os.path.join(folder, 'graph.nnef')
            with open(graph_filename, 'w') as file:
                _print(graph, file, extensions=self._extensions, fragments=fragments,
                       version_custom_ops=self._generate_custom_fragments and self._version_custom_fragments,
                       annotate_shapes=self._annotate_shapes)

            for op in graph.operations:
                if op.type == 'variable':
                    filename = op.attribs['label'] + ".dat"
                    _write_tensor(np.asarray(op.output.data, order='C'), os.path.join(folder, filename),
                                  quantized=True if op.output.quant else False)

            if any(tensor.quant for tensor in graph.tensors):
                quant_filename = os.path.join(folder, 'graph.quant')
                with open(quant_filename, 'w') as file:
                    _write_quantization(graph, file)
        finally:
            if self._compression is not None and folder:
                tgz_compress(folder, path + '.tgz', compression_level=self._compression)
                shutil.rmtree(folder)

    @staticmethod
    def _used_operators(graph, dependencies):
        used = {op.type for op in graph.operations}
        count = len(used)
        changed = True
        while changed:
            for key, deps in six.iteritems(dependencies):
                if key in used:
                    used.update(deps)

            changed = len(used) > count
            count = len(used)

        return used



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import

from .helpers import *
import flatbuffers
try:
    from flatbuffers import flexbuffers
    has_flexbuffers = True
except ImportError:
    has_flexbuffers = False
import numpy as np


def _CreateNumpyVector(builder, x):
    if not isinstance(x, np.ndarray):
        raise TypeError("non-numpy-ndarray passed to CreateNumpyVector")

    if x.dtype.kind not in ['b', 'i', 'u', 'f']:
        raise TypeError("numpy-ndarray holds elements of unsupported datatype")

    if x.ndim > 1:
        raise TypeError("multidimensional-ndarray passed to CreateNumpyVector")

    builder.StartVector(x.itemsize, x.size, x.dtype.alignment)

    # Ensure little endian byte ordering
    if x.dtype.str[0] != "<":
        x = x.byteswap()

    length = x.itemsize * x.size
    builder.head -= length
    builder.Bytes[builder.head: builder.head + length] = x.tobytes()

    return builder.EndVector(x.size)


def _build_buffer(builder, bytes):
    data = _CreateNumpyVector(builder, bytes)
    fb.BufferStart(builder)
    fb.BufferAddData(builder, data)
    return fb.BufferEnd(builder)


def _build_tensor(builder, tensor, buffer_index):
    name = builder.CreateString(tensor.name)
    type = DtypeFromNumpy[tensor.data.dtype.type if isinstance(tensor.data, np.ndarray) else tensor.dtype]

    fb.TensorStartShapeVector(builder, len(tensor.shape))
    for s in reversed(tensor.shape):
        builder.PrependInt32(s)
    shape = builder.EndVector(len(tensor.shape))

    buffer = buffer_index if tensor.data is not None else 0

    quant = _build_quantization(builder, tensor.quant, type)

    fb.TensorStart(builder)
    fb.TensorAddName(builder, name)
    fb.TensorAddShape(builder, shape)
    fb.TensorAddType(builder, type)
    fb.TensorAddBuffer(builder, buffer)
    if quant is not None:
        fb.TensorAddQuantization(builder, quant)
    return fb.TensorEnd(builder)


def _ensure_numpy_array(x, dtype):
    if isinstance(x, np.ndarray):
        assert x.dtype == dtype
        return x
    else:
        return np.array(x, dtype=dtype)


def _build_quantization(builder, quant, dtype):
    if quant is None:
        return None

    min = quant.get('min')
    max = quant.get('max')
    zero_point = quant.get('zero_point')
    scale = quant.get('scale')

    if all(item is None or item == 0 for item in [min, max, zero_point, scale]):
        return None

    min = _CreateNumpyVector(builder, _ensure_numpy_array(min, dtype=np.float32)) if min is not None else None
    max = _CreateNumpyVector(builder, _ensure_numpy_array(max, dtype=np.float32)) if max is not None else None
    scale = _CreateNumpyVector(builder, _ensure_numpy_array(scale, dtype=np.float32)) if scale is not None else None
    zero_point = _CreateNumpyVector(builder, _ensure_numpy_array(zero_point, dtype=np.int64)) if zero_point is not None else None

    fb.QuantizationParametersStart(builder)
    if dtype != fb.TensorType.INT32:
        if min is not None:
            fb.QuantizationParametersAddMin(builder, min)
        if max is not None:
            fb.QuantizationParametersAddMax(builder, max)
    if scale is not None:
        fb.QuantizationParametersAddScale(builder, scale)
    if zero_point is not None:
        fb.QuantizationParametersAddZeroPoint(builder, zero_point)
    return fb.QuantizationParametersEnd(builder)


def _build_operator_code(builder, operation):
    builtinCode = BuiltinOperatorValueByType[operation.type] if not operation.custom else fb.BuiltinOperator.CUSTOM
    customCode = builder.CreateString(operation.type) if operation.custom else None

    fb.OperatorCodeStart(builder)
    fb.OperatorCodeAddBuiltinCode(builder, builtinCode)
    if customCode:
        fb.OperatorCodeAddCustomCode(builder, customCode)
    return fb.OperatorCodeEnd(builder)


def _build_operator_options(builder, attribs, optionsClass):
    starter, ender = get_options_starter_ender(optionsClass)
    adders = enumerate_options_adders(optionsClass)
    vector_starters = enumerate_options_vector_starters(optionsClass)

    vector_values = {}
    for name, vector_starter in vector_starters.items():
        value = attribs[name]
        assert isinstance(value, (list, tuple)) and (len(value) == 0 or isinstance(value[0], int))
        vector_starter(builder, len(value))
        for i in reversed(value):
            builder.PrependInt32(i)
        vector_values[name] = builder.EndVector(len(value))

    starter(builder)
    for name, adder in adders.items():
        if name == 'fused_activation_function' and name not in attribs:
            value = 'NONE'
        else:
            value = attribs[name]
            if isinstance(value, type):
                value = DtypeFromNumpy[value]

        value = vector_values.get(name, value)
        value = substitute_enum_name_with_value(name, value, optionsClass)

        adder(builder, value)

    return ender(builder)


def _encode_custom_options(attribs):
    builder = flexbuffers.Builder()
    builder.MapFromElements(attribs)
    return builder.Finish()


def _build_operator_custom_options(builder, attribs):
    value = _encode_custom_options(attribs) if has_flexbuffers else attribs[CustomOptionsKey]

    fb.OperatorStartCustomOptionsVector(builder, len(value))
    for b in reversed(value):
        builder.PrependUint8(b)
    return builder.EndVector(len(value))


def _build_operator(builder, operation, op_code_index, tensor_index):
    inputs = [tensor_index[tensor] for tensor in operation.inputs]
    fb.OperatorStartInputsVector(builder, len(inputs))
    for input in reversed(inputs):
        builder.PrependInt32(input)
    inputs = builder.EndVector(len(inputs))

    outputs = [tensor_index[tensor] for tensor in operation.outputs]
    fb.OperatorStartOutputsVector(builder, len(outputs))
    for output in reversed(outputs):
        builder.PrependInt32(output)
    outputs = builder.EndVector(len(outputs))

    attribs = {name: value for name, value in operation.attribs.items()}

    optionsType = BuiltinOptionsByOperator[BuiltinOperatorValueByType.get(operation.type, fb.BuiltinOperator.CUSTOM)]
    optionsClass = BuiltinOptionsClasses[optionsType]
    options = _build_operator_options(builder, attribs, optionsClass) if optionsClass is not None else None
    custom_options = _build_operator_custom_options(builder, attribs) if operation.custom else None

    fb.OperatorStart(builder)
    fb.OperatorAddOpcodeIndex(builder, op_code_index[operation.type])
    fb.OperatorAddInputs(builder, inputs)
    fb.OperatorAddOutputs(builder, outputs)
    fb.OperatorAddBuiltinOptionsType(builder, optionsType)

    if options:
        fb.OperatorAddBuiltinOptions(builder, options)

    if custom_options:
        fb.OperatorAddCustomOptions(builder, custom_options)

    return fb.OperatorEnd(builder)


# https://github.com/google/flatbuffers/issues/4814
def FinishWithFileIdentifier(builder, rootTable, fid):
    from flatbuffers import number_types as N
    from flatbuffers import encode

    if fid is None or len(fid) != 4:
        raise Exception('fid must be 4 chars')

    flags = N.Uint8Flags
    prepSize = 4
    builder.Prep(builder.minalign, prepSize + len(fid))
    for i in range(3, -1, -1):
        builder.head = builder.head - flags.bytewidth
        encode.Write(flags.packer_type, builder.Bytes, builder.Head(), ord(fid[i]))

    return builder.Finish(rootTable)


def write_flatbuffers(graph, filename):
    graph.sort()
    builder = flatbuffers.Builder(0)

    fb.BufferStartDataVector(builder, 0)
    data = builder.EndVector(0)
    fb.BufferStart(builder)
    fb.BufferAddData(builder, data)
    buffer = fb.BufferEnd(builder)

    buffers = [buffer]
    for tensor in graph.tensors:
        if tensor.data is not None:
            tensor_data = tensor.data
            if not isinstance(tensor_data, np.ndarray):
                tensor_data = np.array(tensor_data, dtype=tensor.dtype)
            bytes = tensor_data.reshape([-1]).view(np.uint8)
            buffers.append(_build_buffer(builder, bytes))

    fb.ModelStartBuffersVector(builder, len(buffers))
    for buffer in reversed(buffers):
        builder.PrependUOffsetTRelative(buffer)
    buffers = builder.EndVector(len(buffers))

    buffer_index = 1

    tensors = []
    tensor_index = {}
    for tensor in graph.tensors:
        tensor_index[tensor] = len(tensors)
        tensors.append(_build_tensor(builder, tensor, buffer_index))
        if tensor.data is not None:
            buffer_index += 1

    fb.SubGraphStartTensorsVector(builder, len(tensors))
    for tensor in reversed(tensors):
        builder.PrependUOffsetTRelative(tensor)
    tensors = builder.EndVector(len(tensors))

    op_codes = []
    op_code_index = {}
    for operation in graph.operations:
        if operation.type not in op_code_index:
            op_code_index[operation.type] = len(op_codes)
            op_codes.append(_build_operator_code(builder, operation))

    fb.ModelStartOperatorCodesVector(builder, len(op_codes))
    for op_code in reversed(op_codes):
        builder.PrependUOffsetTRelative(op_code)
    op_codes = builder.EndVector(len(op_codes))

    operators = []
    for operation in graph.operations:
        operators.append(_build_operator(builder, operation, op_code_index, tensor_index))

    fb.SubGraphStartOperatorsVector(builder, len(operators))
    for operator in reversed(operators):
        builder.PrependUOffsetTRelative(operator)
    operators = builder.EndVector(len(operators))

    name = builder.CreateString(graph.name) if graph.name is not None else None

    inputs = graph.inputs
    fb.SubGraphStartInputsVector(builder, len(inputs))
    for input in reversed(inputs):
        builder.PrependInt32(tensor_index[input])
    inputs = builder.EndVector(len(inputs))

    outputs = graph.outputs
    fb.SubGraphStartInputsVector(builder, len(outputs))
    for output in reversed(outputs):
        builder.PrependInt32(tensor_index[output])
    outputs = builder.EndVector(len(outputs))

    fb.SubGraphStart(builder)
    if name is not None:
        fb.SubGraphAddName(builder, name)
    fb.SubGraphAddTensors(builder, tensors)
    fb.SubGraphAddOperators(builder, operators)
    fb.SubGraphAddInputs(builder, inputs)
    fb.SubGraphAddOutputs(builder, outputs)
    subgraph = fb.SubGraphEnd(builder)

    fb.ModelStartSubgraphsVector(builder, 1)
    builder.PrependUOffsetTRelative(subgraph)
    subgraphs = builder.EndVector(1)

    fb.ModelStart(builder)
    fb.ModelAddVersion(builder, OUTPUT_SCHEMA_VERSION)
    fb.ModelAddBuffers(builder, buffers)
    fb.ModelAddOperatorCodes(builder, op_codes)
    fb.ModelAddSubgraphs(builder, subgraphs)
    model = fb.ModelEnd(builder)

    FinishWithFileIdentifier(builder, model, OUTPUT_FILE_IDENTIFIER)

    with open(filename, 'wb') as file:
        file.write(builder.Output())


class Writer(object):

    def __call__(self, graph, filename):
        write_flatbuffers(graph, filename)



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import
from .converter import ConverterFromNNEF as _Converter, Transform
from ..model import Tensor, Operation
from ..utils import types
import numpy as np
from nnef.shapes import pool_shape, reduce_shape, deconv_shape


class Converter(_Converter):

    @staticmethod
    def defined_shapes():
        return {
            'lp_pool': pool_shape,
            'lp_reduce': reduce_shape,
            'mean_variance_normalization': lambda input, scale, offset, **kwargs: input,
            'erf': lambda x: x,
        }

    @staticmethod
    def decomposed_operations():
        return ['lstm_step', 'lstm_loop']

    def __init__(self, custom_transforms=None, custom_functions=None, mirror_unsupported=False):
        _Converter.__init__(self, transforms=self.merge_transforms(_Transforms, custom_transforms),
                            functions=custom_functions, mirror_unsupported=mirror_unsupported)

    def __call__(self, graph):
        self.fill_data_in_constants(graph)
        self.convert_variables_to_constants(graph)
        graph = _Converter.__call__(self, graph)
        self._fix_inline_constants(graph)
        return graph

    def _fix_inline_constants(self, graph):
        constants = 0
        for tensor in graph.tensors:
            if tensor.name is None:
                constants += 1
                tensor.name = '$' + str(constants)

    def _make_constant(self, graph, dtype, value, inline):
        return Tensor(graph, dtype=dtype, shape=self._shape_of(value), data=types.to_numpy(value, dtype=dtype))

    def _const_operation(self, output, value):
        Operation(output.graph, type='Constant', inputs=(), outputs=output,
                  attribs={'value': types.to_numpy(value, dtype=output.dtype)})

    def _transform_constant(self, tensor, func):
        if tensor.producer:
            data = func(tensor.producer.attribs['value'] if tensor.producer else tensor.data)
            tensor.shape = data.shape
            tensor.producer.attribs['value'] = data
        else:
            tensor.data = func(tensor.data)
            tensor.shape = tensor.data.shape

    def _squeeze_operation(self, input, output, axes):
        Operation(input.graph, type='Squeeze', inputs=input, outputs=output, attribs={'axes': axes})

    def _unsqueeze_operation(self, input, output, axes):
        Operation(input.graph, type='Unsqueeze', inputs=input, outputs=output, attribs={'axes': axes})

    def _interleave(self, items):
        return [item[0] for item in items] + [item[1] for item in items]

    def squeeze_input(self, tensor, axes):
        return self._pre_squeeze(tensor, axes=axes) if len(axes) else tensor

    def squeeze_output(self, tensor, axes):
        return self._post_squeeze(tensor, axes=axes) if len(axes) else tensor

    def unsqueeze_input(self, tensor, axes):
        return self._pre_unsqueeze(tensor, axes=axes) if len(axes) else tensor

    def unsqueeze_output(self, tensor, axes):
        return self._post_unsqueeze(tensor, axes=axes) if len(axes) else tensor

    def squeeze_vector(self, tensor):
        if self._is_constant(tensor) and len(self._tensor_map[tensor].consumers) == 1:
            self._transform_constant(tensor, lambda data: np.squeeze(data, 0))
            return tensor
        else:
            return self.squeeze_input(tensor, axes=[0])

    def convert_pads(self, padding, truncate=False):
        return self._interleave(padding[2:] if truncate else padding) if padding != [] else None

    def convert_auto_pad(self, padding):
        return "SAME_UPPER" if padding == [] else "NOTSET"

    def convert_output_padding(self, input_shape, filter_shape, output_shape, padding, stride, dilation, groups):
        calculated_shape = deconv_shape(input_shape, filter_shape, padding=padding, stride=stride, dilation=dilation, groups=groups)
        output_padding = [o - c for c, o in zip(calculated_shape[2:], output_shape[2:])]
        return output_padding

    def is_const(self, tensor, value=None):
        return self._is_constant(self._tensor_map[tensor]) and value is None or self.as_const(tensor) == value

    def broadcast(self, tensor, rank):
        return self.unsqueeze_input(tensor, axes=list(range(tensor.rank, rank)))


_Transforms = Converter.unpack_transforms({
    ('external', 'constant'):
        Transform(type=None),
    ('conv', 'deconv'):
        Transform(
            type=('Conv', 'ConvTranspose'),
            defaults={
                'output_shape': None,
            },
            using={
                'transposed': '!_type_ == "deconv"',
                'group': '!groups if groups != 0 else O[0].shape[1] if transposed else I[0].shape[1]',
            },
            cond={
                '!I[2].rank != 0 or (is_const(I[2]) and as_const(I[2]) == 0)': 'bias must be constant 0 or of rank 1',
            },
            inputs=(
                '!I[0]',
                '!I[1]',
                '!squeeze_vector(I[2]) if I[2].rank != 0 else None',
            ),
            outputs='!O[0]',
            attribs={
                'auto_pad': '!convert_auto_pad(padding)',
                'pads': '!convert_pads(padding)',
                'strides': '!stride',
                'dilations': '!dilation',
                'group': '!group',
                'output_shape': '!output_shape if _type_ == "deconv" and output_shape != [] and padding == [] else None',
                'output_padding': '!convert_output_padding(I[0].shape, I[1].shape, output_shape, padding=padding, '
                                  'stride=stride, dilation=dilation, groups=group) '
                                  'if _type_ == "deconv" and output_shape != [] and padding != [] else None',
                'kernel_shape': '!I[1].shape[2:]',
            }
        ),
    ('max_pool', 'avg_pool', 'lp_pool'):
        Transform(
            type=('MaxPool', 'AveragePool', 'LpPool'),
            cond={
                '!size[:2] == [1,1]': 'size must be 1 in batch and channel dimensions',
                '!stride[:2] == [1,1]': 'stride must be 1 in batch and channel dimensions',
                '!dilation[:2] == [1,1]': 'dilation must be 1 in batch and channel dimensions',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'kernel_shape': '!size[2:]',
                'auto_pad': '!convert_auto_pad(padding)',
                'pads': '!convert_pads(padding, truncate=True)',
                'strides': '!stride[2:]',
                'dilations': '!dilation[2:] if _type_ == "max_pool" and not all(d == 1 for d in dilation[2:]) else None',
                'count_include_pad': '!(1 if border == "constant" else 0) if _type_ == "avg_pool" else None',
            }
        ),
    ('min_reduce', 'max_reduce', 'mean_reduce', 'sum_reduce'):
        Transform(
            type=('ReduceMin', 'ReduceMax', 'ReduceMean', 'ReduceSum'),
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axes': '!axes',
                'keepdims': 1,
            }
        ),
    'lp_reduce':
        Transform(
            type='!"ReduceL1" if p == 1 else "ReduceL2"',
            cond={
                '!p == 1 or p == 2': 'p must be 1 or 2',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axes': '!axes',
                'keepdims': 1,
            }
        ),
    ('argmin_reduce', 'argmax_reduce'):
        Transform(
            type=('ArgMin', 'ArgMax'),
            cond={
                '!len(axes) == 1': 'axes must be of length 1',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axis': '!axes[0]',
                'keepdims': True,
            }
        ),
    'batch_normalization':
        Transform(
            type='BatchNormalization',
            inputs=(
                '!I[0]',
                '!squeeze_vector(I[4])',
                '!squeeze_vector(I[3])',
                '!squeeze_vector(I[1])',
                '!squeeze_vector(I[2])',
            ),
            outputs='!O[0]',
            attribs={
                'epsilon': '!epsilon',
                'spatial': '!0 if I[1].rank == I[0].rank else None',
            }
        ),
    ('relu', 'sigmoid', 'tanh', 'softplus', 'selu', 'not', 'copy', 'elu', 'erf', 'abs', 'sign',
     'sin', 'cos', 'tan', 'asin', 'acos', 'atan', 'sinh', 'cosh', 'tanh', 'asinh', 'acosh', 'atanh',
     'exp', 'log', 'neg', 'sqrt', 'ceil', 'floor', 'round'):
        Transform(
            type=('Relu', 'Sigmoid', 'Tanh', 'Softplus', 'Selu', 'Not', 'Identity', 'Elu', 'Erf', 'Abs', 'Sign',
                  'Sin', 'Cos', 'Tan', 'Asin', 'Acos', 'Atan', 'Sinh', 'Cosh', 'Tanh', 'Asinh', 'Acosh', 'Atanh',
                  'Exp', 'Log', 'Neg', 'Sqrt', 'Ceil', 'Floor', 'Round'),
            inputs='!I[0]',
            outputs='!O[0]',
        ),
    ('add', 'sub', 'mul', 'div', 'pow', 'min', 'max', 'and', 'or', 'eq', 'lt', 'gt', 'le', 'ge'):
        Transform(
            type=('Add', 'Sub', 'Mul', 'Div', 'Pow', 'Min', 'Max', 'And', 'Or',
                  'Equal', 'Less', 'Greater', 'LessOrEqual', 'GreaterOrEqual'),
            inputs=(
                '!broadcast(I[0], O[0].rank)',
                '!broadcast(I[1], O[0].rank)',
            ),
            outputs='!O[0]',
        ),
    'sqr':
        Transform(
            type='Mul',
            inputs=('!I[0]', '!I[0]'),
            outputs='!O[0]',
        ),
    'leaky_relu':
        Transform(
            type='LeakyRelu',
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'alpha': '!alpha',
            }
        ),
    'prelu':
        Transform(
            type='PRelu',
            inputs=(
                '!I[0]',
                '!broadcast(I[1], I[0].rank)',
            ),
            outputs='!O[0]',
        ),
    'transpose':
        Transform(
            type='Transpose',
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'perm': '!axes',
            }
        ),
    'reshape':
        Transform(
            type='Reshape',
            inputs=(
                '!I[0]',
                '!as_tensor(shape, np.int64)',
            ),
            outputs='!O[0]',
        ),
    'squeeze':
        Transform(
            type='Squeeze',
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axes': '!axes',
            }
        ),
    'unsqueeze':
        Transform(
            type='Unsqueeze',
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axes': '!axes',
            }
        ),
    'matmul':
        Transform(
            using={
                'transposed': '!transposeA or transposeB',
            },
            type="!'Gemm' if transposed else 'MatMul'",
            inputs=('!I[0]', '!I[1]'),
            outputs='!O[0]',
            attribs={
                'transA': '!int(transposeA) if transposed else None',
                'transB': '!int(transposeB) if transposed else None',
            }
        ),
    'linear':
        Transform(
            type='Gemm',
            inputs=(
                '!I[0]',
                '!I[1]',
                '!squeeze_vector(I[2])',
            ),
            outputs='!O[0]',
            attribs={
                'transA': 0,
                'transB': 1,
            }
        ),
    'local_response_normalization':
        Transform(
            type='LRN',
            cond={
                '!size[0] == 1 and all(s == 1 for s in size[2:])': 'size must be 1 in all non-channel dimensions',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'alpha': '!alpha',
                'beta': '!beta',
                'bias': '!bias',
                'size': '!size[1]',
            }
        ),
    'concat':
        Transform(
            type='Concat',
            inputs=['!I[:]'],
            outputs='!O[0]',
            attribs={
                'axis': '!axis',
            }
        ),
    'split':
        Transform(
            type='Split',
            using={
                'factor': '!I[0].shape[axis] // sum(ratios)',
            },
            inputs='!I[0]',
            outputs=['!O[:]'],
            attribs={
                'axis': '!axis',
                'split': '![r * factor for r in ratios]',
            }
        ),
    'softmax':
        Transform(
            type='Softmax',
            cond={
                '!len(axes) == 1': 'axes must be of length 1',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axis': '!axes[0]',
            }
        ),
    'add_n':
        Transform(
            type='Sum',
            inputs=['!I[:]'],
            outputs='!O[0]',
        ),
    'select':
        Transform(
            type='Where',
            inputs=(
                '!broadcast(I[0], O[0].rank)',
                '!broadcast(I[1], O[0].rank)',
                '!broadcast(I[2], O[0].rank)',
            ),
            outputs='!O[0]',
        ),
    'clamp':
        Transform(
            type='Clip',
            cond={
                '!I[1].rank == 0': 'input a must be of rank 0',
                '!I[2].rank == 0': 'input b must be of rank 0',
            },
            inputs=(
                '!I[0]',
                '!I[1]',
                '!I[2]',
            ),
            outputs='!O[0]',
        ),
    'pad':
        Transform(
            type='Pad',
            inputs=(
                '!I[0]',
                '!as_tensor(convert_pads(padding), np.int64)',
                '!as_tensor(value, np.float32)',
            ),
            outputs='!O[0]',
            attribs={
                'mode': '!"edge" if border == "replicate" else border',
            }
        ),
    'tile':
        Transform(
            type='Tile',
            inputs=(
                '!I[0]',
                '!as_tensor(repeats, np.int64)',
            ),
            outputs='!O[0]',
        ),
    'slice':
        Transform(
            type='Slice',
            inputs=(
                '!I[0]',
                '!as_tensor(begin, np.int64)',
                '!as_tensor(end, np.int64)',
                '!as_tensor(axes, np.int64)',
                '!as_tensor(stride, np.int64)',
            ),
            outputs='!O[0]',
        ),
    ('l1_normalization', 'l2_normalization'):
        Transform(
            type='LpNormalization',
            cond={
                '!len(axes) == 1': 'axes must be of length 1',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axis': '!axes[0]',
                'p': '!1 if _type_ == "l1_normalization" else 2',
            }
        ),
    'mean_variance_normalization':
        Transform(
            type='!"InstanceNormalization" if instance else "MeanVarianceNormalization"',
            using={
                'instance': '!axes == list(range(2, I[0].rank))'
                            ' and I[1].rank == 2 and I[1].shape[0] == 1'
                            ' and I[2].rank == 2 and I[2].shape[0] == 1',
            },
            cond={
                '!is_const(scale, 1.0) if not instance else True':
                    'scale must be 1 if operation does not denote instance normalization',
                '!is_const(offset, 0.0) if not instance else True':
                    'offset must be 0 if operation does not denote instance normalization',
            },
            inputs=(
                '!I[0]',
                '!squeeze_vector(I[1]) if instance else None',
                '!squeeze_vector(I[2]) if instance else None',
            ),
            outputs='!O[0]',
            attribs={
                'axes': '!axes if not instance else None',
                'epsilon': '!epsilon if instance else None',
            }
        ),
    ('nearest_upsample', 'multilinear_upsample'):
        Transform(
            type='Resize',
            using={
                'linear': '!_type_ == "multilinear_upsample"',
            },
            inputs=(
                '!I[0]',
                '!as_tensor([], np.float32)',
                '!as_tensor([1.0, 1.0] + [float(f) for f in factor], np.float32)',
            ),
            outputs='!O[0]',
            attribs={
                'mode': '!"linear" if linear else "nearest"',
                'coordinate_transformation_mode': '!("half_pixel" if method == "symmetric" else'
                                                  ' "asymmetric" if method == "asymmetric" else'
                                                  ' "align_corners") if linear else None',
            }
        ),
    'nearest_downsample':
        Transform(
            type='Resize',
            inputs=(
                '!I[0]',
                '!as_tensor([], np.float32)',
                '!as_tensor([1.0, 1.0] + [1.0 / f for f in factor], np.float32)',
            ),
            outputs='!O[0]',
            attribs={
                'mode': 'nearest',
            }
        ),
    'gather':
        Transform(
            type='Gather',
            inputs=('!I[0]', '!I[1]'),
            outputs='!O[0]',
            attribs={
                'axis': '!axis',
            },
        ),
    'cast':
        Transform(
            type='Cast',
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'to': '!O[0].dtype',
            }
        ),
})



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import numpy as np
from collections.abc import Sequence


# noinspection PyUnresolvedReferences
def as_str(s):
    if sys.version_info[0] >= 3:
        return s.decode('utf-8') if isinstance(s, bytes) else s
    else:
        return s.encode('utf-8') if isinstance(s, unicode) else s


PyTypeFromNumpyDtype = {
    np.float16: float,
    np.float32: float,
    np.float64: float,
    np.int8: int,
    np.int16: int,
    np.int32: int,
    np.int64: int,
    np.uint8: int,
    np.uint16: int,
    np.uint32: int,
    np.uint64: int,
    np.bool_: bool,
    np.str_: str,
}

PyTypeToNumpyDtype = {
    int: np.int32,
    float: np.float32,
    bool: np.bool_,
    str: np.str_,
}


_builtin_type = type


def cast(value, type):
    return _builtin_type(value)(cast(item, type) for item in value) if isinstance(value, Sequence) else type(value)


def from_numpy(array, type=None):
    if type is None:
        type = PyTypeFromNumpyDtype[array.dtype.type]
    return cast(array.tolist(), type)


def to_numpy(value, dtype=None):
    def _item(value):
        return _item(value[0]) if isinstance(value, Sequence) else value

    if dtype is None:
        dtype = PyTypeToNumpyDtype.get(type(_item(value)))
    return np.array(value, dtype=dtype)
